from .citros_db import CitrosDB
from .citros_dict import CitrosDict

class Stat:
    """
    Class for collecting technical statistics about connections from the CitrosDB class.

    Methods `get_stat()` and `print_stat()` return and display, correspondingly, information
    about how many connections and queries were done by all CitrosDB objects existing in the current session.
    The information is recorded only for those CitrosDB that were created with parameter debug_connect = True:

    - `n_pg_connection` - number of postgres connections
    - `n_pg_queries` - number of postgres queries
    - `pg_calls` - the statistics of how many queries to postgres database were done by each method.
    """

    def get_stat(self, format: str = "dict"):
        """
        Return information about connections.

        Parameters
        ----------
        format : {'dict','CitrosDict'}
            The returning format.

        Returns
        -------
        stat: dict or citros_data_analysis.data_access.citros_dict.CitrosDict
        """
        stat = {
            "n_pg_connections": CitrosDB.n_pg_connections,
            "n_pg_queries": CitrosDB.n_pg_queries,
        }
        if format == "CitrosDict":
            stat["pg_calls"] = CitrosDict(CitrosDB.pg_calls)
            stat = CitrosDict(stat)
        else:
            stat["pg_calls"] = CitrosDB.pg_calls
        return stat

    def print_stat(self):
        """
        Print information about connections.
        """
        self.get_stat(format="CitrosDict").print()



from citros.data.analysis import CitrosData
from citros.data.access import CitrosDict

import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import numpy as np
from termcolor import colored
from matplotlib.ticker import MaxNLocator
from typing import Union
from citros.data.access._utils import _get_logger

class Validation:
    """
    Validation class.

    Parameters
    ----------
    df : pandas.DataFrame
        Data table to perform validation tests on.
    data_label : str or list of str
        Specifies the label(s) of the data column(s) in data table.
    param_label : str
        Specifies the label of the column used to calculate the indices.
    method : {'scale', 'bin'}, default 'scale'
        Method of data preparation: scaling to [0,1] interval or binning.
    num : int, default 100
        Number of points in a new scale that will be used for interpolation if the `method` is 'scale'
        or number of bins if the `method` is 'bin'.
    units : str, optional
        Specifies units of the data.
    omit_nan_rows : bool
        If True, rows with one or more NaN values will be omitted from the analysis.
        If not specified, considered to be True.
    inf_vals : None or float, default 1e308
        If specified, all values from `data_label` column that exceed the provided value in absolute terms
        will be treated as NaN values. If this functionality is not required, set inf_vals = None.
    log : log : logging.Logger, default None
        Logger to record log. If None, then the new logger is created.

    Attributes
    ----------
    df : pandas.DataFrame or None
        Data table to perform validation tests on.
    db : analysis.citros_data.CitrosData or None
        CitrosData object after binning or scaling.
    stat : analysis.citros_stat.CitrosStat or None
        CitrosStat object that stores mean, standard deviation and covariance matrix as attributes.

    Examples
    --------
    Import Validation and CitrosDB:

    >>> from citros import CitrosDB, Validation

    From the batch 'albedo' of the simulation 'planetary_nebula' from the json-data column of the topic 'A' 
    download simulated data labeled as 'data.x.x_1' and column with time 'data.time'.

    >>> citros = CitrosDB(simulation = 'planetary_nebula', batch = 'albedo')
    >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x.x_1','data.time'])
    >>> print(df)
        sid   rid   time        topic   type   data.x.x_1   data.time
    0   1     0     312751159   A       a      0.000        10.0
    1   1     1     407264008   A       a      0.008        17.9
    2   1     2     951279608   A       a      0.016        20.3

    Set 'data.time' as independent variable and 'data.x.x_1' as dependent one.
    `method` defines the method of data preparation and index assignment: method = 'bin' - bins values of column `param_label` in `num` intervals,
    set index to each of the interval, group data according to the binning and calculate mean data values for each group.

    >>> V = Validation(df, data_label = ['data.x.x_1'], param_label = 'data.time',
    ...                method = 'bin', num = 50, units = 'm')

    For topic 'A' download 3-dimensional json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns,
    and column with time 'data.time'.
    >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
    >>> print(df['data.x'])
    0          {'x_1': 0.0, 'x_2': 0.08, 'x_3': 0.047}
    1       {'x_1': 0.008, 'x_2': 0.08, 'x_3': -0.003}
    2      {'x_1': 0.016, 'x_2': 0.078, 'x_3': -0.034}
    ...

    Set 'data.time' as independent variable and 'data.x' as dependent vector.
    `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval
    and interpolate data on the new scale.

    >>> V = Validation(df, data_label = 'data.x', param_label = 'data.time',
    ...                method = 'scale', num = 50, units = 'm')
    """

    def __init__(
        self,
        df=None,
        data_label=None,
        param_label=None,
        method="scale",
        num=100,
        units="",
        omit_nan_rows=True,
        inf_vals=1e308,
        log = None,
    ):
        if log is None:
            self.log = _get_logger(__name__)

        self._set_data_table(
            df,
            data_label,
            param_label,
            method=method,
            num=num,
            units=units,
            omit_nan_rows=omit_nan_rows,
            inf_vals=inf_vals,
        )

    def _set_data_table(
        self,
        df,
        data_label,
        param_label,
        method="scale",
        num=100,
        units="",
        omit_nan_rows=True,
        inf_vals=1e308,
    ):
        """
        Set data table to perform validation tests on.

        Parameters
        ----------
        df : pandas.DataFrame
            Data table to perform validation tests on.
        data_label : str or list of str
            Specifies label(s) of the data column(s) in data table.
        param_label : str
            Label of column on the basis of which the indices will be calculated.
        method : {'scale', 'bin'}, default 'scale'
            Method of data preparation: scaling to [0,1] interval or binning.
        num : int, default 100
            Number of points in a new scale that will be used for interpolation or number of bins.
        units : str, optional
            Specifies units of the data.
        omit_nan_rows : bool
            If True, rows with one or more NaN values will be omitted from the analysis.
            If not specified, considered to be True.
        inf_vals : None or float, default 1e308
            If specified, all values from `data_label` column that exceed the provided value in absolute terms
            will be omitted from the analysis. If this functionality is not required, set inf_vals = None.

        See Also
        --------
        analysis.citros_data.CitrosData.bin_data :
            Bin values of column `param_label` in `num` intervals, group data according to the binning and calculate mean values of each group.
        analysis.citros_data.CitrosData.scale_data :
            Scale parameter `param_label` for each of the 'sid' and interpolate data on the new scale.
        """
        if df is not None:
            if isinstance(df, pd.DataFrame):
                self.df = df.copy()
                if data_label is None:
                    self.log.error("`data_label` must be a label of the 'df' column")
                    self.db = None
                    self.stat = None
                    return
                else:
                    dataset = CitrosData(
                        self.df,
                        data_label=data_label,
                        units=units,
                        omit_nan_rows=omit_nan_rows,
                        inf_vals=inf_vals,
                    )
                    if param_label is None:
                        self.log.error("`param_label` must be a label of the 'df' column")
                        self.db = None
                        self.stat = None
                        return
                    else:
                        if method == "bin":
                            self.db = dataset.bin_data(
                                n_bins=num, param_label=param_label, show_fig=False
                            )
                        elif method == "scale":
                            self.db = dataset.scale_data(
                                n_points=num, param_label=param_label, show_fig=False
                            )
                        else:
                            self.log.error("`method` must be 'scale' or 'bin")
                            self.db = None
                            self.stat = None
                            return
                        self.stat = self.db.get_statistics(return_format="citrosStat")
            else:
                self.log.error("`df` must be a pandas DataFrame")
                self.df = None
                self.db = None
                self.stat = None
                return
        else:
            self.df = None
            self.db = None
            self.stat = None

    def std_bound_test(
        self,
        limits: Union[float, list] = 1.0,
        n_std: int = 3,
        nan_passed: bool = True,
        std_color: str = "b",
        connect_nan_std: bool = False,
        std_area: bool = False,
        std_lines: bool = True,
    ):
        """
        Test whether `n_std`-standard deviation boundary is within the given limits.

        The output is:
        - dictionary with summary of the test results, with the following structure:
        ```python
        {
        'test_param' : list,          # initial tests parameters
        column_name:                  # label of the column, str
            {'passed' : bool},        # if the tests was passed or not
            {'pass_rate' : float},    # fraction of the points that pass the test
            {'failed' : 
                {x_index: x_value}},  # indexes and values of the x coordinate of 
                                      #   the points that fail the test {int: float} 
            {'nan_std' :
                {x_index: x_value}}   # indexes and values of the x coordinate of the points
        }                             #   that have NaN (Not a Number) values for standard deviation
        ```
        - table that provides test results for each standard deviation boundary point, 
        using a boolean value to indicate whether it passed or failed the test;

        - figure with plotted simulations, mean values, standard deviation boundaries and limit boundaries;

        Parameters
        ----------
        limits : float or list, default 1.0
            Limit to test standard deviation boundary. Limits may be set as:
            - one value and it will be treated as an +- interval: value -> [-value, value];
            - list of lower and upper limits: [lower_limit, upper_limit];
            - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector with corresponding standard deviation boundaries [std_bound_1, std_bound_2, std_bound_3]:
            [[**limit_lower**, **limit_upper**], **value_1**, **value_2**] will be processed as: 
            **limit_lower** < std_bound_1 < **limit_upper**,
            -**value_1** < std_bound_2 < **value_1**,
            -**value_2** < std_bound_2 < **value_2**.

        n_std : int, default 3
            The parameter specifies the number of standard deviations to be within limits.
        nan_passed : bool, default True
            If True, the NaN values of standard deviation will pass the test.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with validation test results.
        table : pandas.DataFrame
            Table with test results for each of the standard deviation boundary point, indicating whether it passes or fails the test.
        fig : matplotlib.figure.Figure
            Figure with plotted simulations, mean values, standard deviation boundaries and limit boundaries.

        Other Parameters
        ----------------
        std_color : str, default 'b'
            Color for displaying standard deviations, blue by default.
        connect_nan_std : bool, default False
            If True, all non-NaN values in standard deviation boundary line are connected, resulting in a continuous line. 
            Otherwise, breaks are introduced in the standard deviation line whenever NaN values are encountered.
        std_area : bool, default False
            Fill area within `n_std`-standard deviation lines with color.
        std_lines : bool, default True
            If False, remove standard deviation boundary lines.

        See Also
        --------
        pandas.DataFrame, pandas.Series

        Examples
        --------
        Import Validation and CitrosDB:

        >>> from citros import CitrosDB, Validation

        From the batch 'density' of the simulation 'diffuse_nebula' from the topic 'A' download 2 columns of the simulated data 
        labeled 'data.x.x_1' and 'data.x.x_2' and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x.x_1' and 'data.x.x_2' as dependent 2-dimensional vector.
        `method` defines the method of data preparation and index assignment: method = 'bin' - bins values of column `param_label` in `num` intervals, 
        set index to each of the interval, group data according to the binning and calculate mean data values for each group.
        
        >>> citros = CitrosDB(simulation = 'diffuse_nebula', batch = 'density')
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'})\\
        ...                       .data(['data.x.x_1','data.x.x_2','data.time'])
        >>> V = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2'], param_label = 'data.time', 
        ...                method = 'bin', num = 50, units = 'm')

        Test whether 3-sigma standard deviation boundary is within interval [-0.3, 0.3] (treat nan values of the
        standard deviation, if they exist, as passing the test):

        >>> log, table, fig = V.std_bound_test(limits = 0.3, n_std = 3, nan_passed = True)
        >>> log.print()
        std_bound_test: passed
        {
         'test_param': {
           'limits': 0.3,
           'n_std': 3,
           'nan_passed': True
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
         }
        }

        The same, but set limit interval to be [-1, 0.3]:

        >>> log, table, fig = V.std_bound_test(limits = [-1, 0.3], n_std = 3, nan_passed = True)
        std_bound_test: passed

        Set different limits for 1-sigma standard deviation boundaries of 2-dimensional vector: for the first 
        element of the vector boundaries should be within interval [-1, 2] and for the second one - [-0.5, 0.5]:

        >>> log, table, fig = V.std_bound_test(limits = [[-1, 2], 0.5], n_std = 1)
        std_bound_test: passed

        The same as in the previous example, but limits should be [-1, 1] for the first element of the vector 
        and [-0.5, 0.5] for the second. In this case limits should be set as [[-1, 1], [-0.5, 0.5]] and not as [1, 0.5],
        because in the latter case limits will be treated as a common boundary for both elements.

        >>> log, table, fig = V.std_bound_test(limits = [[-1, 1], [-0.5, 0.5]], n_std = 1)
        std_bound_test: passed

        Download 3-dimensional json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns, and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x' as dependent vector.
        `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval 
        and interpolate data on the new scale.
        
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
        >>> V3 = Validation(df, data_label = 'data.x', param_label = 'data.time', 
        ...                 method = 'scale', num = 50, units = 'm')

        Set different limits on 3-dimensional vector: [-0.5, 0.5] for the first element, [-1.5, 1.5] for the second,
        [-20, 10] for the third:

        >>> log, table, fig = V3.std_bound_test(limits = [0.5, 1.5, [-20, 10]], n_std = 3)
        std_bound_test: passed
        """
        fig, ax = self.db._plot_statistics(
            self.stat,
            fig_title="Std boundary test",
            # show_fig=False,
            return_fig=True,
            n_std=n_std,
            std_color=std_color,
            connect_nan_std=connect_nan_std,
            std_area=std_area,
            std_lines=std_lines,
        )
        lower_limit, upper_limit = self._get_limit_values(
            limits, len(self.db.data.columns)
        )
        if lower_limit is None:
            return None, None, None

        nan_std = ~self.stat.sigma.notna()  # True if the value is nan
        if_nan_sigma = nan_std.any()  # if there are any nan values in the column

        valid_b = (self.stat.mean - n_std * self.stat.sigma - lower_limit) > 0
        valid_t = (self.stat.mean + n_std * self.stat.sigma - upper_limit) < 0

        result = valid_t * valid_b.values

        # change results of the test for nan std values if nan_passed is True and the nan value are presented
        if if_nan_sigma.any() and nan_passed:
            result.where(
                ~nan_std, True, inplace=True
            )  # change test results to True if the value was nan

        # insert column with `param_label` as a first column
        result.insert(
            0,
            self.db.x_label,
            self.db.addData.groupby(self.db.xid_label).first()[self.db.x_label],
        )

        # get log
        init_param = {"limits": limits, "n_std": n_std, "nan_passed": nan_passed}
        log = self._get_mean_std_log(init_param, result)

        # adding information about nan values of the standard deviations
        for col in self.db.data.columns:
            if if_nan_sigma[col]:
                log[col]["nan_std"] = CitrosDict(
                    result[self.db.x_label].loc[nan_std[col]].to_dict()
                )
            else:
                log[col]["nan_std"] = CitrosDict()

        # print_result
        all_pass = []
        for col in self.db.data.columns:
            all_pass.append(log[col]["passed"])

        if np.array(all_pass).all():
            print("std_bound_test: " + colored("passed", "green"))
        else:
            print("std_bound_test: " + colored("failed", "red"))

        # plot bounds on fig
        self._plot_bounds(result, lower_limit, upper_limit, fig)

        return log, result, fig

    def mean_test(self, limits: Union[float, list] = 1.0, nan_passed: bool = True):
        """
        Test whether mean is within the given limits.

        The output is:
        - dictionary with summary of the test results, with the following structure:
        ```python
        {
        'test_param' : list,          # initial tests parameters
        column_name:                  # label of the column, str
            {'passed' : bool},        # if the tests was passed or not.
            {'pass_rate' : float},    # fraction of the points that pass the test
            {'failed' : 
                {x_index: x_value}},  # indexes and values of the x coordinate of the 
        }                             #   points that fail the test {int: float}   
        ```

        - table that provides test results for each of the mean point, 
        using a boolean value to indicate whether it passed or failed the test;

        - figure with plotted simulations, mean values and limit boundaries.

        Parameters
        ----------
        limits : float or list, default 1.0
            Limit to test mean. Limits may be set as:
            - one value and it will be treated as an +- interval: value -> [-value, value];
            - list of lower and upper limits: [lower_limit, upper_limit];
            - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector with corresponding mean vector [mean_1, mean_2, mean_3]:
            [[**limit_lower**, **limit_upper**], **value_1**, **value_2**] will be processed as: 
            **limit_lower** < mean_1 < **limit_upper**,
            -**value_1** < mean_2 < **value_1**,
            -**value_2** < mean_2 < **value_2**.

        nan_passed : bool, default True
            If True, the NaN values of the mean will pass the test.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with validation test results.
        table : pandas.DataFrame
            Table with test results for each of the mean point, indicating whether it passes or fails the test.
        fig : matplotlib.figure.Figure
            Figure with plotted simulations, mean values and limit boundaries.
        
        Examples
        --------
        Import Validation and CitrosDB:

        >>> from citros import CitrosDB, Validation

        From the batch 'density' of the simulation 'diffuse_nebula' from the topic 'A' download 2 columns of the simulated data 
        labeled 'data.x.x_1' and 'data.x.x_2' and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x.x_1' and 'data.x.x_2' as dependent 2-dimensional vector.
        `method` defines the method of data preparation and index assignment: method = 'bin' - bins values of column `param_label` in `num` intervals, 
        set index to each of the interval, group data according to the binning and calculate mean data values for each group.
        
        >>> citros = CitrosDB(simulation = 'diffuse_nebula', batch = 'density')
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'})\\
        ...                       .data(['data.x.x_1','data.x.x_2','data.time'])
        >>> V = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2'], param_label = 'data.time', 
        ...                method = 'bin', num = 50, units = 'm')

        Test whether mean values are is within the  interval [-10, 10]:

        >>> log, table, fig = V.mean_test(limits = 10)
        >>> log.print()
        mean_test: passed
        {
         'test_param': {
           'limits': 10
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           }
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           }
         }
        }

        The same, but set limit interval to be [-0.5, 0.8]:

        >>> log, table, fig = V.mean_test(limits = [-0.5, 0.8])
        mean_test: passed

        Set different limits on mean values for each of the 1-dimensional element of the 2-dimensional vector: 
        [-0.05, 0.08] for the first element and [-0.5, 0.5] for the second:

        >>> log, table, fig = V.mean_test(limits = [[-0.05, 0.08], 0.5])
        mean_test: passed

        The same as in the previous example, but limits should be [-1, 1] for the first element of the vector 
        and [-0.5, 0.5] for the second. In this case limits should be set as [[-1, 1], [-0.5, 0.5]] and not as [1, 0.5],
        because in the latter case limits will be treated as a common boundary for both elements.

        >>> log, table, fig = V.mean_test(limits = [[-1, 1], [-0.5, 0.5]])
        mean_test: passed

        Download 3-dimensional json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns, and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x' as dependent vector.
        `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval 
        and interpolate data on the new scale.
        
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
        >>> V3 = Validation(df, data_label = 'data.x', param_label = 'data.time', 
        ...                 method = 'scale', num = 50, units = 'm')

        Set different limits on 3-dimensional vector: [-0.5, 0.5] for the first element, [-1.5, 1.5] for the second,
        [-20, 10] for the third:
        
        >>> log, table, fig = V3.mean_test(limits = [0.5, 1.5, [-20, 10]])
        mean_test: passed
        """
        fig, ax = self.db._plot_statistics(
            self.stat,
            fig_title="Mean test",
            # show_fig=False,
            return_fig=True,
            n_std=None,
            std_color="b",
        )
        lower_limit, upper_limit = self._get_limit_values(
            limits, len(self.db.data.columns)
        )
        if lower_limit is None:
            return None, None, None

        nan_mean = ~self.stat.mean.notna()  # True if the value is nan
        if_nan_mean = nan_mean.any()  # if there are any nan values in the column

        valid_b = (self.stat.mean - lower_limit) > 0
        valid_t = (self.stat.mean - upper_limit) < 0

        result = valid_t * valid_b.values

        # change results of the test for nan mean values if nan_passed is True and the nan value are presented
        if if_nan_mean.any() and nan_passed:
            result.where(
                ~nan_mean, True, inplace=True
            )  # change test results to True if the value was nan

        # insert column with `param_label` as a first column
        result.insert(
            0,
            self.db.x_label,
            self.db.addData.groupby(self.db.xid_label).first()[self.db.x_label],
        )

        # get log and put plot bounds on fig
        init_param = {"limits": limits}
        log = self._get_mean_std_log(init_param, result)

        # print results
        all_pass = []
        for col in self.db.data.columns:
            all_pass.append(log[col]["passed"])

        if np.array(all_pass).all():
            print("mean_test: " + colored("passed", "green"))
        else:
            print("mean_test: " + colored("failed", "red"))

        # plot bounds on fig
        self._plot_bounds(result, lower_limit, upper_limit, fig)

        return log, result, fig

    def std_test(
        self,
        limits: Union[float, list] = 1.0,
        n_std: int = 3,
        nan_passed: bool = True,
        std_color: str = "b",
        connect_nan_std: bool = False,
        std_area: bool = False,
        std_lines: bool = True,
    ):
        """
        Test whether `n_std`-standard deviation does not exceed the given limits.

        The output is:
        - dictionary with summary of the test results, with the following structure:
        ```python
        {
        'test_param' : list,          # initial tests parameters
        column_name:                  # label of the column, str
            {'passed' : bool},        # if the tests was passed or not
            {'pass_rate' : float},    # fraction of the points that pass the test
            {'failed' : 
                {x_index: x_value}},  # indexes and values of the x coordinate of 
                                      #   the points that fail the test {int: float} 
            {'nan_std' :
                {x_index: x_value}}   # indexes and values of the x coordinate of the points
        }                             #   that have NaN (Not a Number) values for standard deviation
        ```

        - table that provides test results for each standard deviation point, 
        using a boolean value to indicate whether it passed or failed the test.

        - figure with standard deviations and limit boundaries.

        Parameters
        ----------
        limits : float or list, default 1.0
            Limit to test standard deviation. Limits may be set as:
            - one value;
            - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. 
            For example, in case of 3-dimensional vector with corresponding standard deviation vector = [std_1, std_2, std_3], limits = [**value_1**, **value_2**, **value_3**] will be processed as:
            std_1 < **value_1**,
            std_2 < **value_2**,
            std_2 < **value_3**.

        n_std : int, default 3
            The parameter specifies the number of standard deviations to be less then limits.
        nan_passed : bool, default True
            If True, the NaN values of standard deviation will pass the test.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with validation test results.
        table : pandas.DataFrame
            Table with test results for each of the standard deviation point, indicating whether it passes or fails the test.
        fig : matplotlib.figure.Figure
            Figure with standard deviations and limit boundaries.

        Other Parameters
        ----------------
        std_color : str, default 'b'
            Color for displaying standard deviation, blue by default.
        connect_nan_std : bool, default False
            If True, all non-NaN values in standard deviation line are connected, resulting in a continuous line. 
            Otherwise, breaks are introduced in the standard deviation line whenever NaN values are encountered.
        std_area : bool, default False
            Fill area within `n_std`-standard deviation line with color.
        std_lines : bool, default True
            If False, remove standard deviation line.

        See Also
        --------
        pandas.DataFrame, pandas.Series

        Examples
        --------
        Import Validation and CitrosDB:

        >>> from citros import CitrosDB, Validation

        From the batch 'density' of the simulation 'diffuse_nebula' from the topic 'A' download 2 columns of the simulated data 
        labeled 'data.x.x_1' and 'data.x.x_2' and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x.x_1' and 'data.x.x_2' as dependent 2-dimensional vector.
        `method` defines the method of data preparation and index assignment: method = 'bin' - bins values of column `param_label` in `num` intervals, 
        set index to each of the interval, group data according to the binning and calculate mean data values for each group.
        
        >>> citros = CitrosDB(simulation = 'diffuse_nebula', batch = 'density')
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'})\\
        ...                       .data(['data.x.x_1','data.x.x_2','data.time'])
        >>> V = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2'], param_label = 'data.time', 
        ...                method = 'bin', num = 50, units = 'm')

        Test whether 3-sigma standard deviation is within interval [-0.3, 0.3] (treat nan values of the
        standard deviation, if they exist, as passing the test):

        >>> log, table, fig = V.std_test(limits = 1.5, n_std = 3, nan_passed = True)
        >>> log.print()
        std_bound_test: passed
        {
         'test_param': {
           'limits': 0.3,
           'n_std': 3,
           'nan_passed': True
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
         }
        }

        Download 3-dimensional json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns, and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x' as dependent vector.
        `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval 
        and interpolate data on the new scale.
        
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
        >>> V3 = Validation(df, data_label = 'data.x', param_label = 'data.time', 
        ...                 method = 'scale', num = 50, units = 'm')

        Set different limits on 3-dimensional vector: 1.5 for the first element, 1.5 for the second,
        30 for the third:

        >>> log, table, fig = V3.std_test(limits = [1.5, 1.5, 30], n_std = 3)
        std_test: passed
        """
        upper_limit = self._get_1D_limit(limits, len(self.db.data.columns))
        if upper_limit is None:
            return None, None, None

        nan_std = ~self.stat.sigma.notna()  # True if the value is nan
        if_nan_sigma = nan_std.any()  # if there are any nan values in the column

        result = (n_std * self.stat.sigma - upper_limit) < 0

        # change results of the test for nan std values if nan_passed is True and the nan value are presented
        if if_nan_sigma.any() and nan_passed:
            result.where(
                ~nan_std, True, inplace=True
            )  # change test results to True if the value was nan

        # insert column with `param_label` as a first column
        result.insert(
            0,
            self.db.x_label,
            self.db.addData.groupby(self.db.xid_label).first()[self.db.x_label],
        )

        # get log
        init_param = {"limits": limits, "n_std": n_std, "nan_passed": nan_passed}
        log = self._get_mean_std_log(init_param, result)

        # adding information about nan values of the standard deviations
        for col in self.db.data.columns:
            if if_nan_sigma[col]:
                log[col]["nan_std"] = CitrosDict(
                    result[self.db.x_label].loc[nan_std[col]].to_dict()
                )
            else:
                log[col]["nan_std"] = CitrosDict()

        # print_result
        all_pass = []
        for col in self.db.data.columns:
            all_pass.append(log[col]["passed"])

        if np.array(all_pass).all():
            print("std_test: " + colored("passed", "green"))
        else:
            print("std_test: " + colored("failed", "red"))

        # plot fig
        fig, ax = plt.subplots(nrows=len(self.db.data.columns), ncols=1)
        if not isinstance(ax, np.ndarray):
            ax = np.array([ax])
        if len(upper_limit.shape) == 0:
            upper_limit = [upper_limit] * len(self.stat.sigma.columns)

        for i, y_col in enumerate(self.stat.sigma.columns):
            res_sigm = self.stat.sigma[y_col]
            y_0 = pd.Series([0] * len(res_sigm), index=res_sigm.index)
            if connect_nan_std:
                filter_st = res_sigm.notna()
                line_style_custom = "-"
            else:
                filter_st = pd.Series([True] * len(res_sigm), index=res_sigm.index)
                if res_sigm.notna().all():
                    line_style_custom = "-"
                else:
                    line_style_custom = ".-"
            if std_lines:
                ax[i].plot(
                    result[self.db.x_label][filter_st],
                    n_std * res_sigm[filter_st],
                    line_style_custom,
                    color=std_color,
                    label=str(n_std) + r"$\sigma$",
                    markersize=0.7,
                )
                label = None
            else:
                label = str(n_std) + r"$\sigma$"
            if std_area:
                ax[i].fill_between(
                    result[self.db.x_label][filter_st].values,
                    n_std * res_sigm[filter_st].values,
                    y_0[filter_st].values,
                    color=std_color,
                    label=label,
                    alpha=0.12,
                )

            ax[i].plot(
                [min(result[self.db.x_label]), max(result[self.db.x_label])],
                [upper_limit[i], upper_limit[i]],
                "r",
                label="limit",
            )
            ylabel = (
                "std " + y_col + ", [" + self.db.units + "]"
                if self.db.units != ""
                else ""
            )
            ax[i].set_ylabel(ylabel)
            ax[i].grid(True)
        fig.supxlabel(self.db.x_label)
        fig.suptitle("Std test")
        handles, labels = ax[-1].get_legend_handles_labels()
        fig.legend(handles, labels, bbox_to_anchor=(1.0, 0.94), loc="upper left")
        fig.tight_layout()
        # self._plot_bounds(result, self.stat.mean - upper_limit, self.stat.mean + upper_limit, fig)

        return log, result, fig

    def sid_test(self, limits: Union[float, list] = 1.0, nan_passed: bool = True):
        """
        Test whether all simulations are within the given limits.

        The output is:
        - dictionary with summary of the test results, with the following structure:
        ```python
        {
        'test_param' : list,                # initial tests parameters
        column_name:                        # label of the column, str
            {'passed' : bool},              # if the tests was passed or not.
            {'pass_rate' : 
                {'sid_fraction' : float},   # fraction of simulations that pass the test
                {sid : fraction}},          # fraction of the points that pass the test for each simulation {int: float}
            {'failed' : 
                {sid :                      # id of the simulation that contains points that failed the test
                    {x_index: x_value}}},   # indexes and values of the x coordinate of the points 
        }                                   #   that fail the test {int: {int: float}}
        ```

        - table that provides test results for for each point of the simulations, 
        using a boolean value to indicate whether it passed or failed the test;

        - figure with plotted simulations, mean values and limit boundaries.

        Parameters
        ----------
        limits : float or list, default 1.0
            Limit to test simulation results. Limits may be set as:
            - one value and it will be treated as an +- interval: value -> [-value, value];
            - list of lower and upper limits: [lower_limit, upper_limit];
            - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector that contains v1, v2, v3 columns and numbers N simulations:
            [[**limit_lower**, **limit_upper**], **value_1**, **value_2**] will be processed as: 
            **limit_lower** < v1 < **limit_upper**, 
            -**value_1** < v2 < **value_1**, 
            -**value_2** < v3 < **value_2** for each of the N simulations.

        nan_passed : bool, default True
            If True, the NaN values will pass the test.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with validation test results.
        table : pandas.DataFrame
            Table with test results for each point of the simulations, indicating whether it passes or fails the test.
        fig : matplotlib.figure.Figure
            Figure with plotted simulations, mean values and limit boundaries.

        Examples
        --------
        Import Validation and CitrosDB:

        >>> from citros import CitrosDB, Validation

        From the batch 'density' of the simulation 'diffuse_nebula' from the topic 'A' download 2 columns of the simulated data 
        labeled 'data.x.x_1' and 'data.x.x_2' and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x.x_1' and 'data.x.x_2' as dependent 2-dimensional vector.
        `method` defines the method of data preparation and index assignment: method = 'bin' - bins values of column `param_label` in `num` intervals, 
        set index to each of the interval, group data according to the binning and calculate mean data values for each group.
        
        >>> citros = CitrosDB(simulation = 'diffuse_nebula', batch = 'density')
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'})\\
                                  .data(['data.x.x_1','data.x.x_2','data.time'])
        >>> V = Validation(df, data_label = ['data.x.x_1', 'data.x.x_2'], param_label = 'data.time', 
        ...                method = 'bin', num = 50, units = 'm')

        Test whether all simulations are is within the interval [-10, 10]:

        >>> log, table, fig = V.sid_test(limits = 10)
        >>> log.print()
        sid_test: passed
        {
         'test_param': {
           'limits': 10
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           }
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           }
         }
        }

        The same, but set limit interval to be [-0.5, 0.8]:

        >>> log, table, fig = V.sid_test(limits = [-0.5, 0.8])
        sid_test: passed

        Set different limits on mean values for each of the 1-dimensional element of the 2-dimensional vector: 
        [-0.05, 0.08] for the first element and [-0.5, 0.5] for the second:

        >>> log, table, fig = V.sid_test(limits = [[-0.05, 0.08], 0.5])
        sid_test: passed

        The same as in the previous example, but limits should be [-1, 1] for the first element of the vector 
        and [-0.5, 0.5] for the second. In this case limits should be set as [[-1, 1], [-0.5, 0.5]] and not as [1, 0.5],
        because in the latter case limits will be treated as a common boundary for both elements.

        >>> log, table, fig = V.sid_test(limits = [[-1, 1], [-0.5, 0.5]])
        sid_test: passed

        For topic 'A' download 3-dimensional json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns, and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x' as dependent vector.
        `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval 
        and interpolate data on the new scale.
        
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
        >>> V3 = Validation(df, data_label = 'data.x', param_label = 'data.time', 
        ...                 method = 'scale', num = 50, units = 'm')

        Set different limits on 3-dimensional vector: [-0.5, 0.5] for the first element, [-1.5, 1.5] for the second one, an
        [-20, 10] for the third vector element:
        
        >>> log, table, fig = V3.sid_test(limits = [0.5, 1.5, [-20, 10]])
        sid_test: passed
        """
        fig, ax = self.db._plot_statistics(
            self.stat,
            fig_title="Sid test",
            # show_fig=False,
            return_fig=True,
            n_std=None,
            std_color="b",
        )

        lower_limit, upper_limit = self._get_limit_values(
            limits, len(self.db.data.columns)
        )
        if lower_limit is None:
            return None, None, None

        nan_val = ~self.db.data.notna()  # True if the value is nan
        if_nan_val = nan_val.any()  # if there are any nan values in the column

        valid_b = (self.db.data - lower_limit) > 0
        valid_t = (self.db.data - upper_limit) < 0

        result = valid_t * valid_b.values

        # change results of the test for nan values if nan_passed is True and the nan value are presented
        if if_nan_val.any() and nan_passed:
            result.where(
                ~nan_val, True, inplace=True
            )  # change test results to True if the value was nan

        # insert column with `param_label` as a first column
        result.insert(0, self.db.x_label, self.db.addData[self.db.x_label])

        try:
            pass_rate = (
                (
                    result.groupby("sid")
                    .agg({col: "value_counts" for col in self.db.data.columns})
                    .xs(True, level=1)
                    / result[self.db.data.columns].groupby("sid").count()
                )
                .fillna(0)
                .apply(round, ndigits=3)
                .to_dict()
            )
        except KeyError:
            # in case all points failed test and there is no True values
            pass_rate = CitrosDict()
            for col in self.db.data.columns:
                sid = list(set(result[col].index.get_level_values("sid")))
                sid.sort()
                pass_rate[col] = CitrosDict(dict.fromkeys(sid, 0.0))

        try:
            pass_rate_overall = (
                (
                    result.groupby("sid")
                    .all()
                    .agg({col: "value_counts" for col in self.db.data.columns})
                    .loc[True]
                    / result[self.db.data.columns].groupby("sid").all().count().values
                )
                .fillna(0)
                .apply(round, ndigits=3)
                .to_dict()
            )
        except KeyError:
            # in case all points failed test and there is no True values
            pass_rate_overall = CitrosDict()
            for col in self.db.data.columns:
                pass_rate_overall[col] = 0.0

        passed = result.groupby("sid")[self.db.data.columns].all().all()

        log = CitrosDict()

        log["test_param"] = CitrosDict({"limits": limits})

        for col in self.db.data.columns:
            log[col] = CitrosDict()
            log[col]["passed"] = passed[col]
            log[col]["pass_rate"] = CitrosDict()
            log[col]["pass_rate"]["sid_fraction"] = pass_rate_overall[col]
            log[col]["pass_rate"].update(CitrosDict(pass_rate[col]))
            if log[col]["passed"]:
                log[col]["failed"] = CitrosDict()
            else:
                log[col]["failed"] = CitrosDict(
                    result[result[col] == False]
                    .swaplevel()
                    .groupby("sid")
                    .apply(
                        lambda f: CitrosDict(f.xs(f.name)[self.db.x_label].to_dict())
                    )
                    .to_dict()
                )
            # log[col]['table'] = result[[col]]

        # print results
        all_pass = []
        for col in self.db.data.columns:
            all_pass.append(log[col]["passed"])

        if np.array(all_pass).all():
            print("sid_test: " + colored("passed", "green"))
        else:
            print("sid_test: " + colored("failed", "red"))

        # plot bounds on fig
        self._plot_bounds(result, lower_limit, upper_limit, fig)

        return log, result, fig

    def norm_test(self, norm_type: str = "L2", limits: Union[float, list] = 1.0):
        """
        Test whether norm of the each simulation is less than the given limit.

        The output is:
        - dictionary with summary of the test results, with the following structure:
        ```python
        {
        'test_param' : list,                # initial tests parameters
        column_name :                       # label of the column, str
            {'passed' : bool},              # if the tests was passed or not.
            {'pass_rate' : float}           # fraction of the simulations that pass the test
            {'norm_value' :
                {sid: value}},              # norm for each of the simulation {int: float}
            {'failed' : list}               # sid that fail the test
        }
        ```

        - table that provides test results for each simulation,
        using a boolean value to indicate whether it passed or failed the test;

        - figure with plotted norm value and limits.

        Parameters
        ----------
        norm_type : {'L2', 'Linf'}, default 'L2'
            Norm type. Norm is calculated for each of the simulation. If data is a multidimensional vector, it is calculated
            for each simulation of the each vector element.
            Type of the norm:
            - 'L2' - Euclidean norm, square root of the sum of the squares.
            - 'Linf' - absolute maximum.

        limits : float or list, default 1.0
            Limits on the simulation norm. Limits may be set as:
            - one value;
            - if the data has multiple columns, limits may be set for each of the column separately as a list.
            That way list length must be equal to number of the columns.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with validation test results.
        table : pandas.DataFrame
            Table with test results for each simulation, indicating whether it passes or fails the test.
        fig : matplotlib.figure.Figure
            Figure with plotted norm value and limits.

        Examples
        --------
        Import Validation and CitrosDB:

        >>> from citros import CitrosDB, Validation

        From the batch 'density' of the simulation 'diffuse_nebula' from the topic 'A' download 1 columns of the simulated data 
        labeled 'data.x.x_1' and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x.x_1' as a dependent one.
        `method` defines the method of data preparation and index assignment: method = 'bin' - bins values of column `param_label` in `num` intervals,
        set index to each of the interval, group data according to the binning and calculate mean data values for each group.

        >>> citros = CitrosDB(simulation = 'diffuse_nebula', batch = 'density')
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x.x_1','data.time'])
        >>> V = Validation(df, data_label = 'data.x.x_1', param_label = 'data.time',
        ...                method = 'bin', num = 50, units = 'm')

        Test whether L2 norm for each of the simulation does not exceed 1:

        >>> log, table, fig = V.norm_test(norm_type = 'L2', limits = 1)
        >>> log.print()
        >>> print(table)
        norm_test L2: passed
        {
         'test_param': {
           'limits': 1
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'norm_value': {
             1: 0.39,
             2: 0.39,
             3: 0.38
           },
           'failed': []
         },
        }
        >>> print(table)
             data.x.x_1
        sid
        1          True
        2          True
        3          True

        Download 3-dimensional json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns, and column with time 'data.time'.
        Set 'data.time' as independent variable and 'data.x' as dependent vector.
        `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval
        and interpolate data on the new scale.

        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
        >>> V3 = Validation(df, data_label = 'data.x', param_label = 'data.time',
        ...                method = 'scale', num = 50, units = 'm')

        Set different limits on Linf norm for each of the element of the 3-dimensional vector: 1.0 for the first element,
        0.1 for the second one, and 0.5 for the third vector element:

        >>> log, table, fig = V3.norm_test(norm_type = 'Linf', limits = [1.0, 0.1, 0.5])
        norm_test Linf: passed
        """
        upper_limit = self._get_1D_limit(limits, len(self.db.data.columns))
        if upper_limit is None:
            return None, None, None

        norm = self._get_norm(norm_type)
        if norm is None:
            return None, None, None

        result = (norm - upper_limit) < 0

        try:
            pass_rate = (
                (
                    result.agg(
                        {col: "value_counts" for col in self.db.data.columns}
                    ).loc[True]
                    / result.count().values
                )
                .fillna(0)
                .apply(round, ndigits=3)
                .to_dict()
            )
        except KeyError:
            pass_rate = CitrosDict()
            for col in self.db.data.columns:
                pass_rate[col] = 0.0

        log = CitrosDict()
        log["test_param"] = CitrosDict({"limits": limits})

        for col in self.db.data.columns:
            log[col] = CitrosDict()
            log[col]["passed"] = result.all()[col]
            log[col]["pass_rate"] = pass_rate[col]
            log[col]["norm_value"] = CitrosDict(norm[col].to_dict())
            log[col]["failed"] = [k for k, v in result[col].to_dict().items() if not v]

        all_pass = []
        for col in self.db.data.columns:
            all_pass.append(log[col]["passed"])

        if np.array(all_pass).all():
            print("norm_test " + norm_type + ": " + colored("passed", "green"))
        else:
            print("norm_test " + norm_type + ": " + colored("failed", "red"))

        fig, axes = plt.subplots(nrows=len(norm.columns), ncols=1, figsize=(6, 6))
        if not isinstance(axes, np.ndarray):
            axes = [axes]

        for i, ax in enumerate(axes):
            col = norm.columns[i]
            index = norm[col].index
            ax.barh(index, norm[col].values, align="center", label="sid norm")

            ax.yaxis.set_major_locator(MaxNLocator(integer=True))

            y_min = min(index)
            y_max = max(index)
            y_tick_labels = []
            y_tick_position = []
            for y_tick in fig.axes[0].get_yticklabels():
                y_pos = y_tick.get_position()[1]
                if (y_pos >= y_min) and (y_pos <= y_max):
                    y_tick_labels.append(y_tick.get_text())
                    y_tick_position.append(y_pos)

            if len(y_tick_labels) > 1:
                if str(y_min) not in y_tick_labels:
                    y_tick_labels = [str(y_min)] + y_tick_labels
                    y_tick_position = [y_min] + y_tick_position
                if str(y_max) not in y_tick_labels:
                    y_tick_labels = y_tick_labels + [str(y_max)]
                    y_tick_position = y_tick_position + [y_max]
            ax.set_yticks(y_tick_position, labels=y_tick_labels)

            ylabel = col
            ax.set_ylabel(ylabel)
            ylim = ax.get_ylim()
            if isinstance(upper_limit, np.ndarray) and len(upper_limit.shape) != 0:
                ax.plot(
                    [upper_limit[i], upper_limit[i]],
                    [min(index) - 1, max(index) + 1],
                    "r",
                    label="limit",
                )
            else:
                ax.plot(
                    [upper_limit, upper_limit],
                    [min(index) - 1, max(index) + 1],
                    "r",
                    label="limit",
                )
            ax.set_ylim(ylim)
            ax.xaxis.grid(True)

        handles, labels = axes[-1].get_legend_handles_labels()
        fig.legend(handles, labels, bbox_to_anchor=(1.0, 0.9), loc="upper left")
        label = {"L2": r"$L^2$", "Linf": r"$L^{inf}$"}
        fig.supylabel("sid")
        supxlabel = ", [" + self.db.units + "]" if self.db.units != "" else ""
        fig.supxlabel("norm" + supxlabel)
        fig.suptitle("Norm: " + label[norm_type] + "\nsid vs. norm")
        fig.tight_layout()

        return log, result, fig

    def set_tests(
        self,
        test_method: dict = {
            "std_bound": {"limits": 1.0, "n_std": 3, "nan_passed": True},
            "mean": {"limits": 1.0, "nan_passed": True},
            "sid": {"limits": 1.0, "nan_passed": True},
            "norm_L2": {"limits": 1.0},
            "norm_Linf": {"limits": 1.0},
        },
    ):
        """
        Perform tests on the data.

        Possible test methods are:
        <details>
            <summary>'std_bound'</summary>

        Test whether standard deviation is within the given limits.
        Test parameters are stored as the dict with the following keys:
        - 'limits' : float or list, default 1.0
            Limit to test standard deviation boundary. Limits may be set as:
           - one value and it will be treated as an +- interval: value -> [-value, value];
           - list of lower and upper limits: [lower_limit, upper_limit];
           - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector with corresponding standard deviation boundaries [std_bound_1, std_bound_2, std_bound_3]:
            [[**limit_lower**, **limit_upper**], **value_1**, **value_2**] will be processed as:
            **limit_lower** < std_bound_1 < **limit_upper**,
            -**value_1** < std_bound_2 < **value_1**,
            -**value_2** < std_bound_2 < **value_2**.
        - 'n_std' : int, default 3
            The parameter specifies the number of standard deviations to be within limits.
        - 'nan_passed' : bool, default True
            If True, the NaN values of standard deviation will pass the test.

        </details>

        <details>
            <summary>'mean'</summary>

        Test whether mean is within the given limits.
        Test parameters are stored as the dict:
        - 'limits' : float or list, default 1.0
            Limit to test mean. Limits may be set as:
           - one value and it will be treated as an +- interval: value -> [-value, value];
           - list of lower and upper limits: [lower_limit, upper_limit];
           - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector
            with corresponding mean vector [mean_1, mean_2 and mean_3]:
            [[**limit_lower**, **limit_upper**], **value_1**, **value_2**] will be processed as:
            **limit_lower** < mean_1 < **limit_upper**,
            -**value_1** < mean_2 < **value_1**,
            -**value_2** < mean_2 < **value_2**.
        - 'nan_passed' : bool, default True
            If True, the NaN values of the mean will pass the test.

        </details>

        <details>
            <summary>'std'</summary>

        Test whether standard deviation is less then the given limits.
        Test parameters are stored as the dict:
        - 'limits' : float or list, default 1.0
            Limit to test standard deviation. Limits may be set as:
           - one value;
           - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector
            with corresponding standard deviation vectors [std_1, std_2, std_3]:
            limits = [**value_1**, **value_2**, **value_3**] will be processed as:
            std_1 < **value_1**,
            std_2 < **value_2**,
            std_2 < **value_3**.
        - 'n_std' : int, default 3
            The parameter specifies the number of standard deviations to be less then limits.
        - 'nan_passed' : bool, default True
            If True, the NaN values of the mean will pass the test.

        </details>

        <details>
            <summary>'sid'</summary>

        Test whether all simulations are within the given limits.
        Test parameters are stored as the dict:
        - 'limits' : float or list, default 1.0
            Limit to test simulation results. Limits may be set as:
           - one value and it will be treated as an +- interval: value -> [-value, value];
           - list of lower and upper limits: [lower_limit, upper_limit];
           - If the data has multiple columns, limits may be set for each of the column.
            That way list length must be equal to number of columns. For example, for the 3-dimensional vector that
            contains v1, v2, v3 columns and numbers N simulations:
            [[**limit_lower**, **limit_upper**], **value_1**, **value_2**] will be processed as:
            **limit_lower** < v1 < **limit_upper**,
            -**value_1** < v2 < **value_1**,
            -**value_2** < v3 < **value_2** for each of the N simulations.
        - 'nan_passed' : bool, default True
            If True, the NaN values will pass the test.

        </details>

        <details>
            <summary>'norm_L2'</summary>

        Test whether L2 norm of the each simulation is less than the given limit.
        Test parameters are stored as the dict:
        - 'limits' : float or list, default 1.0
            Limits on the simulation norm. Limits may be set as:
           - one value;
           - if the data has multiple columns, limits may be set for each of the column separately as a list.
            That way list length must be equal to number of the columns.

        </details>

        <details>
            <summary>'norm_Linf'</summary>

        Test whether Linf norm of the each simulation is less than the given limit.
        Test parameters are stored as the dict:
        - 'limits' : float or list, default 1.0
            Limits on the simulation norm. Limits may be set as:
           - one value;
           - if the data has multiple columns, limits may be set for each of the column separately as a list.
            That way list length must be equal to number of the columns.

        </details>

        Parameters
        ----------
        tests_method : dict
            Keys define test methods and corresponding test parameters are stored as values.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with the test results.
        tables : dict
            Dictionary with test methods as keys and pandas.DataFrame table with results of the test as values.
        figures : dict
            Dictionary with test methods as keys and matplotlib.figure.Figure with test results as values.

        See Also
        --------
        Validation.std_bound_test, Validation.mean_test, Validation.std_test, Validation.sid_test, Validation.norm_test

        Examples
        --------
        Import Validation and CitrosDB:

        >>> from citros import CitrosDB, Validation

        From the batch 'density' of the simulation 'diffuse_nebula' from the topic 'A' download 3-dimensional 
        json-data 'data.x' that contains 'data.x.x_1', 'data.x.x_2' and 'data.x.x_3' columns,
        and column with time 'data.time'.

        >>> citros = CitrosDB(simulation = 'diffuse_nebula', batch = 'density')
        >>> df = citros.topic('A').set_order({'sid':'asc','rid':'asc'}).data(['data.x','data.time'])
        >>> print(df['data.x'])
        0          {'x_1': 0.0, 'x_2': 0.08, 'x_3': 0.047}
        1       {'x_1': 0.008, 'x_2': 0.08, 'x_3': -0.003}
        2      {'x_1': 0.016, 'x_2': 0.078, 'x_3': -0.034}
        ...

        Set 'data.time' as independent variable and 'data.x' as dependent vector.
        `method` defines the method of data preparation and index assignment: method = 'scale' - scales parameter `param_label` for each of the 'sid' to [0, 1] interval
        and interpolate data on the new scale.

        >>> V = Validation(df, data_label = 'data.x', param_label = 'data.time',
        ...                method = 'scale', num = 50, units = 'm')

        Test whether 3 standard deviation boundary is within [-0.3, 0.3] interval (treat nan values of the
        standard deviation, if they are presented, as passed the test) and L2 norm of the each simulation is less than 12.5:

        >>> logs, tables, figs = V.set_tests(test_method = {
        ...                                    'std_bound' : {'limits' : 0.3, 'n_std': 3, 'nan_passed': True},
        ...                                    'norm_L2' : {'limits' : 12.5}})
        std_bound_test: passed
        norm_test L2: passed

        Print detailed standard deviation boundary test results:

        >>> logs['std_bound'].print()
        {
         'test_param': {
           'limits': 0.3,
           'n_std': 3,
           'nan_passed': True
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'failed': {
           },
           'nan_std': {
             49: 807.942
           }
          }
        }

        Print results of norm test in details:

        >>> logs['norm_L2'].print()
        {
         'test_param': {
           'limits': 12.5
         },
         'data.x.x_1': {
           'passed': True,
           'pass_rate': 1.0,
           'norm_value': {
             1: 0.39,
             2: 0.38,
             3: 0.38
           },
           'failed': []
         },
         'data.x.x_2': {
           'passed': True,
           'pass_rate': 1.0,
           'norm_value': {
             1: 0.38,
             2: 0.40,
             3: 0.40
           },
           'failed': []
         },
         'data.x.x_3': {
           'passed': True,
           'pass_rate': 1.0,
           'norm_value': {
             1: 0.12,
             2: 0.11,
             3: 0.12
           },
           'failed': []
         }
        }
        """
        methods = test_method.keys()

        logs = CitrosDict()
        figures = {}
        tables = {}

        bound_tests = {
            "std_bound": self.std_bound_test,
            "mean": self.mean_test,
            "sid": self.sid_test,
            "std": self.std_test,
        }
        norm_tests = {"norm_L2": "L2", "norm_Linf": "Linf"}

        for method in methods:
            if method in bound_tests.keys():
                logs[method], tables[method], figures[method] = bound_tests[method](
                    **test_method[method]
                )
            elif method in norm_tests.keys():
                logs[method], tables[method], figures[method] = self.norm_test(
                    norm_type=norm_tests[method], **test_method[method]
                )

        return logs, tables, figures    
    
    def _get_norm(self, norm_type="L2"):
        """
        Calculates norm.

        Parameters
        ----------
        norm_type : {'L2', 'Linf'}, default 'L2'
            Norm type.
        """
        if norm_type == "L2":
            norm = (
                self.db.data.apply(lambda x: x**2)
                .groupby("sid")
                .sum()
                .apply(lambda x: np.sqrt(x))
            )
        elif norm_type == "Linf":
            norm = (abs(self.db.data)).groupby("sid").max()
        else:
            self.log.error("Can not recognize the norm type, the allowed types are 'L2' and 'Linf'")
            norm = None
        return norm

    def _get_mean_std_log(self, init_param, result):
        """
        Write log for std and mean bound tests.

        Parameters
        ----------
        init_param : dict
            Initial parameters to write in the log.
        result : pandas.DataFrame
            Table with results of the test.

        Returns
        -------
        log : access.citros_dict.CitrosDict
            Dictionary with information for each column: 'passed' (True or False), 'pass_rate', 'failed' (indexes and x values of the
            points that fail the test).
        """
        log = CitrosDict()

        # init parameters
        log["test_param"] = CitrosDict(init_param)

        # calculate pass rate (how many points pass)
        try:
            pass_rate = CitrosDict(
                (
                    result[self.db.data.columns].apply(pd.Series.value_counts).loc[True]
                    / len(result)
                )
                .fillna(0)
                .apply(round, ndigits=3)
            )
        except KeyError:
            # in case all points failed test and there is no True values
            pass_rate = CitrosDict()
            for col in self.db.data.columns:
                pass_rate[col] = 0.0

        for col in self.db.data.columns:
            log[col] = CitrosDict()
            log[col]["passed"] = True if result[col].all() else False
            log[col]["pass_rate"] = pass_rate[col]
            log[col]["failed"] = CitrosDict(
                result[self.db.x_label]
                .loc[result.index[result[col] == False]]
                .to_dict()
            )
            # log[col]['table'] = result[[col]]

        return log

    def _plot_bounds(self, result, lower_limit, upper_limit, fig):
        """
        Add bound lines on the plot.

        Parameters
        ----------
        result : pandas.DataFrame
            Table with results of the test.
        lower_limit : numpy.ndarray
            Lower limit(s).
        upper_limit : numpy.ndarray
            Upper limit(s).
        fig : matplotlib.figure.Figure
            Figure to plot bound lines on.
        """
        axes = fig.axes
        for i, ax in enumerate(axes):
            if isinstance(lower_limit, np.ndarray) and len(lower_limit.shape) != 0:
                ax.plot(
                    [min(result[self.db.x_label]), max(result[self.db.x_label])],
                    [lower_limit[i], lower_limit[i]],
                    "r",
                )
            elif isinstance(lower_limit, pd.DataFrame) and len(lower_limit.shape) != 0:
                ax.plot(result[self.db.x_label], lower_limit, "r")
            else:
                ax.plot(
                    [min(result[self.db.x_label]), max(result[self.db.x_label])],
                    [lower_limit, lower_limit],
                    "r",
                )

            if isinstance(upper_limit, np.ndarray) and len(upper_limit.shape) != 0:
                ax.plot(
                    [min(result[self.db.x_label]), max(result[self.db.x_label])],
                    [upper_limit[i], upper_limit[i]],
                    "r",
                )
            elif isinstance(upper_limit, pd.DataFrame) and len(upper_limit.shape) != 0:
                ax.plot(result[self.db.x_label], upper_limit, "r")
            else:
                ax.plot(
                    [min(result[self.db.x_label]), max(result[self.db.x_label])],
                    [upper_limit, upper_limit],
                    "r",
                )

        line = Line2D([0], [0], label="test bounds", color="r")
        handles, labels = axes[-1].get_legend_handles_labels()
        fig.legend(
            handles + [line],
            labels + ["test bounds"],
            bbox_to_anchor=(1.0, 0.94),
            loc="upper left",
        )
        fig.legends.pop(0)
        fig.tight_layout()
 
    def _get_limit_values(self, limits, data_length):
        """
        Get lower and upper limits from `limits`.

        Parameters
        ----------
        limits : int or float or list or tuple.
        data_length : int
            Length of the data.

        Returns
        -------
        lower_limit : numpy.ndarray
            Lower limit(s).
        upper_limit : numpy.ndarray
            Upper limit(s).
        """
        if isinstance(limits, (list, tuple)):
            # [0.1]  [0.1, 0.2]  [(0.1, 0.2), (0.1, 0.2)]  [0.1, (0.1, 0.2)]

            if len(limits) == 1:
                # [0.1]
                try:
                    lower_limit = -limits[0]
                    upper_limit = limits[0]
                except TypeError:
                    self.log.error("Could not resolve the limits")
                    return None, None

            elif len(limits) == 2 and not any(
                map(lambda x: isinstance(x, (list, tuple)), limits)
            ):
                # [0.1, 0.2]
                lower_limit = limits[0]
                upper_limit = limits[1]

            else:
                if len(limits) != data_length:
                    self.log.error("`limits` length does not match the number of data columns.")
                    return None, None
                else:
                    # [(0.1, 0.2), (0.1, 0.2)]  [0.1, (0.1, 0.2)]  [0.1, 0.2, 0.3]  [(0.1, 0.2), 0.1, (10, 100)]  [(0.1, 0.2), (0.1, 0.2), (10, 100)]
                    lower_limit = []
                    upper_limit = []

                    for i in range(data_length):
                        if isinstance(limits[i], (list, tuple)) and len(limits[i]) != 1:
                            # [(0.1, 0.2),...
                            lower_limit.append(limits[i][0])
                            upper_limit.append(limits[i][1])
                        elif isinstance(limits[i], (list, tuple)):
                            # [(0.1),...
                            lower_limit.append(-limits[i][0])
                            upper_limit.append(limits[i][0])
                        else:
                            # [0.1,...
                            lower_limit.append(-limits[i])
                            upper_limit.append(limits[i])
        else:
            # single +-value, int or float
            lower_limit = -limits
            upper_limit = limits

        return np.array(lower_limit), np.array(upper_limit)

    def _get_1D_limit(self, limits, data_length):
        """
        Get limit on norm from `limits`.

        Parameters
        ----------
        limits : int or float or list or tuple.
        data_length : int
            Length of the data.

        Returns
        -------
        norm_limit : numpy.ndarray
            Limit(s).
        """
        if isinstance(limits, (list, tuple)):
            # [0.1]  [0.1, 0.2, 0.3]

            if len(limits) == 1:
                # [0.1]
                norm_limit = limits[0]
            else:
                if len(limits) != data_length:
                    self.log.error("`limits` length does not match the number of data columns.")
                    return None
                else:
                    # [0.1, 0.2, 0.3]
                    norm_limit = limits
        else:
            # single value, int or float
            norm_limit = limits

        return np.array(norm_limit)



import sys
import logging
from rich.logging import RichHandler
from rich import print, inspect, print_json
from logging.handlers import TimedRotatingFileHandler


FORMATTER = logging.Formatter(
    "[%(asctime)s] [%(name)s.%(funcName)s:%(lineno)d] [%(levelname)s]: %(message)s"
)


def get_rich_hangler():
    handler = RichHandler(rich_tracebacks=True)
    handler.setFormatter(FORMATTER)
    return RichHandler(rich_tracebacks=True)


# log to console
def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


# log to file
def get_file_handler(log_file):
    file_handler = TimedRotatingFileHandler(log_file, when="midnight")
    file_handler.setFormatter(FORMATTER)
    return file_handler


def str_to_log_level(level: str):
    if level == "debug" or level == "DEBUG":
        return logging.DEBUG
    elif level == "info" or level == "INFO":
        return logging.INFO
    elif level == "warn" or level == "WARN":
        return logging.WARNING
    elif level == "error" or level == "ERROR":
        return logging.ERROR
    else:
        # default to debug on invalid input
        return logging.INFO


def get_logger(
    logger_name,
    log_level="info",
    log_file=".citros/logs/citros.log",
    verbose=False,
):
    logger = logging.getLogger(logger_name)
    logger.setLevel(str_to_log_level(log_level))

    file_handler = get_file_handler(log_file)
    logger.addHandler(file_handler)

    # write to console when debugging
    if verbose:
        # logger.addHandler(get_console_handler())
        logger.addHandler(get_rich_hangler())

    return logger


def shutdown_log():
    logging.shutdown()



import os
import json
import sys
import uuid
import shutil
import importlib_resources

from pathlib import Path
from citros.stats import SystemStatsRecorder
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich_argparse import RichHelpFormatter
from datetime import datetime

from .logger import get_logger, shutdown_log

from .utils import suppress_ros_lan_traffic, validate_dir
from .parameter_setup import ParameterSetup
from .citros_obj import (
    CitrosObj,
    CitrosException,
    CitrosNotFoundException,
    FileNotFoundException,
    NoValidException,
)
from .events import EventsOTLP


class Simulation(CitrosObj):
    """Object representing .citros/simulations/name.json file."""

    ##################
    ##### public #####
    ##################
    def _get_simulation_run_log(self, simulation_rec_dir):
        return get_logger(
            __name__,
            log_level=os.environ.get("LOGLEVEL", "DEBUG" if self.debug else "INFO"),
            log_file=str(simulation_rec_dir / "citros.log"),
            verbose=self.verbose,
        )

    def _copy_ros_log(self, destination: str):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._copy_ros_log()")
        from .utils import get_last_created_file, copy_files, rename_file

        ros_logs_dir_path = get_last_created_file(
            Path("~/.ros/log/").expanduser(), dirs=True
        )
        if get_last_created_file is None:
            self.log.warning(f"Failed to find the ros logs directory.")
            return

        log_file_path = Path(ros_logs_dir_path, "launch.log")
        copy_files([log_file_path], destination, self.log)
        new_file_path = Path(destination, log_file_path.name)
        rename_file(new_file_path, "ros.log")

    @staticmethod
    def _guess_msgtype(path: Path) -> str:
        """
        Guess the message type based on the path.

        Args:
            path (Path): The file path to the message type.

        Returns:
            str: The guessed message type.
        """
        name = path.relative_to(path.parents[2]).with_suffix("")
        if "msg" not in name.parts:
            name = name.parent / "msg" / name.name
        return str(name)

    def _register_custom_message(self, msgpath: str):
        """
        Register custom message types from a given folder path.

        Args:
            message_folder_path (str): The folder path where custom messages are stored.
        """
        from rosbags.typesys import get_types_from_msg, register_types

        msgdef = msgpath.read_text(encoding="utf-8")
        add_types = get_types_from_msg(msgdef, self._guess_msgtype(msgpath))
        self.log.debug(f"{'   '*self.level}regisering {msgpath} ...")

        register_types(add_types)

    def _copy_msg_files(self, destination: str):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.copy_msg_files()")
        from .utils import copy_files
        from .parsers import ParserRos2

        msg_paths = ParserRos2(self.log).get_msg_files(self.root)
        for msg_path in msg_paths:
            # assuming msg files are under package_name/msg/
            package_name = Path(msg_path).parent.parent.name
            target_dir = Path(destination, package_name, "msg")
            copy_files([msg_path], str(target_dir), self.log, True)

            # register custom messages
            self._register_custom_message(str(msg_path))

    def _handle_msg_files(self, simulation_rec_dir):
        self._copy_msg_files(simulation_rec_dir)

    def _save_system_vars(self, destination: str):
        import subprocess
        from os import linesep

        self.log.debug(
            f"{'   '*self.level}{self.__class__.__name__}._save_system_vars()"
        )
        # Get all environment variables
        env_vars = dict(os.environ)

        pip_freeze_output = subprocess.run(
            ["pip", "freeze"], capture_output=True, text=True
        )

        if pip_freeze_output.returncode != 0:
            self.log.error("pip freeze failed: " + pip_freeze_output.stderr)
            python_packages = []
        else:
            python_packages = pip_freeze_output.stdout.split(linesep)

        data = {"environment_variables": env_vars, "python_packages": python_packages}

        with open(Path(destination, "environment.json"), "w") as f:
            json.dump(data, f, indent=4)

    @staticmethod
    def _add_string_connection(writer, connection, msgtype):
        """
        Adds a new connection with a modified topic name by appending '/path' to the original topic.

        This method is specifically intended for handling string connections. After modifying the topic name,
        it uses the provided writer to add the new connection, keeping other connection parameters unchanged.

        Args:
            writer (Writer): An object responsible for writing data.
            connection (Connection): The original connection object.
            msgtype (str): The message type for the new connection.

        Returns:
            list: A list containing the writer and the newly created connection object.

        """
        from typing import TYPE_CHECKING, cast
        from rosbags.interfaces import ConnectionExtRosbag2, Connection

        ext = cast(ConnectionExtRosbag2, connection.ext)
        new_connection = writer.add_connection(
            topic=connection.topic + "/path",
            msgtype=msgtype,
            serialization_format=ext.serialization_format,
            offered_qos_profiles=ext.offered_qos_profiles,
        )
        return writer, new_connection

    @staticmethod
    def _create_string_topic(writer, file_name, connection):
        """
        Creates a new topic with a string message type, intended for writing a file path.

        This method attempts to create a new connection using the provided writer and an existing connection.
        The topic of the new connection is the original topic name with '/path' appended to it.
        It then creates a new String message populated with the given file name.

        Args:
            writer (Writer): The writer object responsible for writing data.
            file_name (str): The file name to be used in the new String message.
            connection (Connection): The original connection object.

        Returns:
            tuple: A tuple containing the writer, the new connection object, the new data (String message),
                and the message type for the new connection.

        Raises:
            Exception: Catches any exceptions thrown during the process. Handles specifically the case when
                    a connection with the new topic name already exists.
        """
        from rosbags.serde import deserialize_cdr, serialize_cdr
        from rosbags.rosbag2 import Reader, Writer
        from rosbags.interfaces import ConnectionExtRosbag2, Connection

        msgtype = String.__msgtype__
        try:
            # try to add string topic to the bag
            writer, new_connection = Simulation._add_string_connection(
                writer, connection, msgtype
            )
        except Exception as e:
            if "Connection can only be added once" in str(e):
                # the connection is already created, take it from the connection list
                new_connection = [
                    x
                    for x in writer.connections
                    if x.topic == connection.topic + "/path"
                ][0]
        new_data = String(file_name)
        return writer, new_connection, new_data, msgtype

    def _create_citros_bag(self, src: str, dst: str):
        """
        Create a new ROS bag file by reading from an existing bag and transforming its contents.

        This method reads messages from an existing ROS bag located at `src`, and writes to a new ROS bag at `dst`.
        It performs custom handling for messages of the type "sensor_msgs/msg/Image" by saving them as image files
        and then writing a new corresponding string message in the new bag.

        Args:
            src (str): The path to the source ROS bag file to be read.
            dst (str): The path to the destination ROS bag file to be created.

        Raises:
            Various exceptions could be raised during reading and writing of the bag files,
            which will be logged using the logger.

        Side Effects:
            - A new ROS bag is created at the location specified by `dst`.
            - Additional image data files and metadata might be created based on the source bag's content.
            - Logging information is produced to inform about the operations and potential errors.

        Note:
            - If an image topic is encountered, the image data is saved separately and a string message is added to the new bag.

        """
        self.log.debug(
            f"{'   '*self.level}{self.__class__.__name__}._create_citros_bag(src={src}, dst={dst})"
        )

        from typing import TYPE_CHECKING, cast
        from rosbags.serde import deserialize_cdr, serialize_cdr
        from rosbags.rosbag2 import Reader, Writer
        from rosbags.interfaces import ConnectionExtRosbag2, Connection

        if src.endswith("/"):
            src = src[:-1]

        # check id name of file contain .citros
        if ".citros" in src.split("/")[-1]:
            self.log.warning(f"CITROS bag already exist, skipping ...")
            return

        self.log.info(f"Creating a new CITROS bag: src = {src}\ndst = {dst}")

        try:
            with Reader(src) as reader, Writer(dst) as writer:
                # create connection map to the writer
                conn_map = {}
                for conn in reader.connections:
                    try:
                        ext = cast(ConnectionExtRosbag2, conn.ext)
                        conn_map[conn.id] = writer.add_connection(
                            topic=conn.topic,
                            msgtype=conn.msgtype,
                            serialization_format=ext.serialization_format,
                            offered_qos_profiles=ext.offered_qos_profiles,
                        )
                    except Exception as e:
                        self.log.error(e)
                        continue

                # reader.connections is a list, we need start from 0 and not from 1 (this is why is x-1)
                rconns = [reader.connections[x - 1] for x in conn_map]

                for connection, timestamp, data in reader.messages(connections=rconns):
                    if "sensor_msgs/msg/Image" in connection.msgtype:
                        # if this is a image topic export data and continue without saving in the new bag.
                        file_name = self._handle_image_message(dst, data, connection)
                        (
                            writer,
                            new_connection,
                            new_data,
                            msgtype,
                        ) = self._create_string_topic(writer, file_name, connection)
                        writer.write(
                            new_connection, timestamp, serialize_cdr(new_data, msgtype)
                        )
                        self.log.info(
                            f"Found image, saving image data under {file_name}"
                        )
                        continue
                    writer.write(conn_map[connection.id], timestamp, data)
                return
        except Exception as e:
            if "citros exists already" in str(e):
                self.log.warning("CITROS bag already exist ...")
                return

    def _get_bags(self, path: str) -> dict:
        """
        Retrieve the paths of bag files in the specified directory and its subdirectories.

        This method recursively traverses the directory structure from the specified path, looking for files with
        ".mcap" or ".db3" extensions. It logs the discovery of any such files.

        Args:
            path (str): The root directory where the search for bag files will commence.

        Returns:
            list: A dict of paths where ".mcap" or ".db3" files are found.

        Raises:
            Exception: If no ".mcap" or ".db3" files are found in the specified path and its subdirectories.

        Note:
            Logging is done to debug the discovery of ".mcap" and ".db3" files.
        """
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._get_bags({path})")
        bags = {}
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.endswith(".mcap"):
                    self.log.debug(f"Found MCAP bag: {file}")
                elif file.endswith(".db3"):
                    self.log.debug(f"Found SQLITE3 bag: {file}")
                else:
                    continue
                bags[root] = file

        if len(bags) < 1:
            raise Exception(
                f"Didn't find SQLITE3 or MCAP bag in the [{path}] folder ..."
            )

        return bags

    def _prepare_citros_bag(self, destination: str):
        # assuming bags already there (the simulations is done recording)
        bags = self._get_bags(f"{destination}/bags/")
        for bag_path, file_name_with_ext in bags.items():
            file_name = os.path.splitext(file_name_with_ext)[0]
            self._create_citros_bag(bag_path, bag_path + f"{file_name}.citros/")

    def run(self, simulation_rec_dir, sid=0, trace_context=None, ros_domain_id=None):
        """Run simulation."""
        # create .citros/data if not exists
        simulation_rec_dir.mkdir(parents=True, exist_ok=True)

        self.log = self._get_simulation_run_log(simulation_rec_dir)

        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.run()")
        events = EventsOTLP(self, trace_context)

        # running inside ROS workspace context.
        from launch import LaunchService
        from citros.ros import generate_launch_description

        self.log.info(f"running simulation [{self.name}]")
        # print(f"running simulation [{self.name}]")

        if self.verbose:
            self.log.info(f'simulation run dir = "{simulation_rec_dir}]"')
        else:
            self.log.debug(f'simulation run dir = "{simulation_rec_dir}]"')

        if ros_domain_id:
            suppress_ros_lan_traffic(ros_domain_id)

        # launch
        launch_description = generate_launch_description(
            self, simulation_rec_dir, sid, events
        )

        if launch_description is None:
            msg = f"ERROR. Failed to create launch_description."
            self.log.error(msg)
            return

        launch_service = LaunchService(debug=False)  # self.debug)
        launch_service.include_launch_description(launch_description)

        systemStatsRecorder = SystemStatsRecorder(f"{simulation_rec_dir}/stats.csv")
        systemStatsRecorder.start()

        # run simulation
        ret = launch_service.run()

        systemStatsRecorder.stop()

        print(
            f"[{'blue' if ret == 0 else 'red'}]Finished simulation with return code [{ret}].",
        )

        try:
            print(f"Copying ros logs...")
            self._copy_ros_log(simulation_rec_dir)
        except Exception as e:
            self.log.error(e)
        try:
            print(f"Copying and registering messages logs...")
            self._handle_msg_files(simulation_rec_dir)
        except Exception as e:
            self.log.error(e)
        try:
            print(f"Saving system vars...")
            self._save_system_vars(simulation_rec_dir)
        except Exception as e:
            self.log.error(e)
        try:
            print(f"Preparing citros bags...")
            self._prepare_citros_bag(simulation_rec_dir)
        except Exception as e:
            self.log.error(e)

        if ret != 0:
            events.error(
                message=f"Finished simulation. Return code = [{ret}].",
            )
            events.on_shutdown()
            # sys.exit(ret)
        else:
            events.done(
                message=f"Finished simulation. Return code = [{ret}].",
            )
        return ret

    # overriding
    def path(self):
        return self.root_citros / "simulations" / f"{self.file}"

    ###################
    ##### private #####
    ###################

    def __init__(
        self,
        name,
        root=None,
        new=False,
        log=None,
        citros=None,
        package_name=None,
        launch_file=None,
        verbose=False,
        debug=False,
        level=0,
    ):
        # used for new simulation
        self.package_name = package_name
        self.launch_file = launch_file
        self.parameter_setup = None
        super().__init__(name, root, new, log, citros, verbose, debug, level)

    # overriding
    def _validate(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._validate()")

        """Validate simulation.json file."""
        Path(self.root).mkdir(parents=True, exist_ok=True)

        # validate json schema is correct
        success = validate_dir(
            self.root_citros / "simulations", "schema_simulation.json", self.log
        )

        if not success:
            self.log.debug(f"{'   '*self.level}{self.__class__.__name__}: False")
            return False

        # validate parameter_setup file
        param_setup = self["parameter_setup"]

        self.parameter_setup = ParameterSetup(
            param_setup,
            root=self.root,
            new=self.new,
            log=self.log,
            citros=self.citros,
            verbose=self.verbose,
            debug=self.debug,
            level=self.level + 1,
        )

        # validate launch file
        launch_file = self["launch"]["file"]
        all_launch_names = [launch["name"] for launch in self._get_launches()]
        if launch_file not in all_launch_names:
            print(
                f'[red]Could not find launch file named {launch_file} referenced in "{self.path()}."',
            )
            self.log.debug(f"{'   '*self.level}{self.__class__.__name__}: return False")
            return False

        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}: return True")
        return True

    # overriding
    def _load(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._load()")

        try:
            # loads self.path()
            super()._load()
        except FileNotFoundError as ex:
            self.log.error(f"simulation file {self.file} does not exist.")
            raise FileNotFoundException(f"simulation file {self.file} does not exist.")

    # overriding
    def _new(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._new()")
        path = self.path()

        # avoid overwrite
        if path.exists():
            self._load()
            # return

        Path(self.root_citros / "simulations").mkdir(parents=True, exist_ok=True)

        # got from __init__
        if self.launch_file is None or self.package_name is None:
            raise ValueError(
                "package_name and launch_file must be provided when creating a new simulation"
            )

        default = {
            "id": str(uuid.uuid4()),
            "description": "Default simulation. Change the values according to your needs.",
            "parameter_setup": "default_param_setup.json",
            "launch": {"file": self.launch_file, "package": self.package_name},
            "timeout": 60,
            "GPU": 0,
            "CPU": 2,
            "MEM": 265,
            "storage_type": "MCAP",
        }
        # self.data = default | self.data
        self.data.update(default)
        self._save()

    ###################
    ##### utils #####
    ###################

    def _get_launches(self):
        """returns a list of launch objects

        Args:
            proj_json (Path): path to project.json file

        Returns:
            [{
                package: str,
                name: str
            }]: array of launch info
        """

        launch_info = []

        for package in self.citros.get("packages", []):
            for launch in package.get("launches", []):
                if "name" in launch:
                    launch_info.append(
                        {"package": package.get("name", ""), "name": launch["name"]}
                    )

        return launch_info

    def _check_simualtion_run_name(self, name):
        batch_name_idx = 1

        if not name or not self.utils.is_valid_file_name(name):
            name = self.utils.get_foramtted_datetime()

        # avoid duplicate batch dir names
        elif Path(
            self._router()["citros"]["runs"]["path"],
            self._sim_name,
            name,
        ).exists():
            while Path(
                self._router()["citros"]["runs"]["path"],
                self._sim_name,
                f"{name}_{str(batch_name_idx)}",
            ).exists():
                batch_name_idx = batch_name_idx + 1
            name = f"{name}_{str(batch_name_idx)}"
        return name



import json
import yaml
import numpy
import shutil
import inspect
import importlib.util
import importlib_resources

from pathlib import Path

from .citros_obj import CitrosObj
from .utils import get_data_file_path, validate_dir
from citros.parsers import ParserRos2


class ParameterSetup(CitrosObj):
    """Object representing .citros/simulation.json file."""

    ##################
    ##### public #####
    ##################

    # overriding
    def path(self):
        return self.root_citros / "parameter_setups" / f"{self.file}"

    def render(self, destination, context={}):
        """
        Fetches parameters from CITROS, saves them to files, and returns the config.

        destination: the simulation recording folder.
        """
        self.log.debug(f"{self.__class__.__name__}._validate()")
        rendered_parameters = self._evaluate(context)

        self.log.debug("Saving parameters to files. ")
        self._save_rendered(rendered_parameters, destination)
        self.log.debug("Done saving config files.")

        return rendered_parameters

    ###################
    ##### private #####
    ###################

    # overriding
    def _validate(self):
        """Validate parameter_setup.json file."""
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._validate()")

        success = validate_dir(
            self.root_citros / "parameter_setups", "schema_param_setup.json", self.log
        )

        return success

    # overriding
    def _new(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._new()")
        path = self.path()

        # avoid overwrite
        if path.exists():
            self._load()
            # return

        Path(self.root_citros / "parameter_setups").mkdir(parents=True, exist_ok=True)

        self.parser_ros2 = ParserRos2(self.log)
        # get default parameter setup and write it to file.
        default_parameter_setup = self.parser_ros2.generate_default_params_setup(
            self.citros
        )
        # self.data = default_parameter_setup | self.data
        self.data.update(default_parameter_setup)
        self._save()

        # TODO[critical]: when addeing a parameter to a package/node, add/append it to the parameter setups that use it.

    def _evaluate(self, context=None):
        """
        evalues self
        self.data = {
            "c": {
                "function": "/path/to/user_defined_function.py:user_function",
                "args": ["a", "b"]
            },
            "b": {
                "function": "numpy.add",
                "args": ["a", 3]
            },
            "a": 5
        }

        print(result)  # Output: {"c": ..., "b": 8, "a": 5}
        """
        self.log.debug(f"{self.__class__.__name__}._evaluate(context={context})")

        def load_function(function_path, function_name):
            self.log.debug(
                f"function_path = {function_path}, function_name = {function_name}"
            )

            if not function_path.startswith("numpy"):
                self.log.debug("Loading user defined function.")
                spec = importlib.util.spec_from_file_location(
                    "user_module", function_path
                )
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                function = getattr(module, function_name)
            else:
                self.log.debug("Loading numpy function.")
                function = eval(function_path)
            return function

        def recursive_key_search(d, key):
            """
            Recursively search for a key within nested dictionaries.

            :param d: The dictionary to search.
            :param key: The key to search for. Can be multi-level e.g. "outer.inner".
            :return: The value corresponding to the given key or None if not found.
            """
            # If the key exists at this level of dictionary, return it
            if key in d:
                return d[key]

            # If key does not contain a ".", search for it recursively within
            # all nested dictionaries
            if "." not in key:
                for k, v in d.items():
                    if isinstance(v, dict):
                        item = recursive_key_search(v, key)
                        if item is not None:
                            return item

            # If key contains a ".", split it and navigate the nested
            # dictionaries accordingly.
            else:
                first, remainder = key.split(".", 1)
                if first in d and isinstance(d[first], dict):
                    return recursive_key_search(d[first], remainder)
                return None

        def collect_all_keys(dct):
            """
            Collect all keys from a nested dictionary.

            :param dct: The dictionary to extract keys from.
            :return: A set containing all keys in the dictionary, including nested keys.
            """
            keys = set(dct.keys())
            for value in dct.values():
                if isinstance(value, dict):
                    keys |= collect_all_keys(value)
            return keys

        def is_valid_reference_key(key):
            """
            Check if a given reference key is valid within a dictionary structure.

            :param key: The reference key to check. Can be multi-level e.g. "outer.inner".
            :return: True if valid, False otherwise.
            """
            parts = key.split(".")
            current_dict = self.data
            for part in parts:
                if part not in current_dict:
                    return False
                current_dict = current_dict[part]
            return True

        all_keys = collect_all_keys(self.data)
        visited_keys = set()  # For circular dependency checks

        def evaluate_value(value):
            """
            Recursively evaluate a value. If the value corresponds to a function,
            it loads and executes the function. If the value is a dictionary or a list,
            it recursively evaluates its contents. If the value is a string that references
            another key, it fetches the value for that key.

            :param value: The value to evaluate.
            :return: The evaluated result.
            """

            # the value is a dictionary representing a function to be executed
            if isinstance(value, dict) and "function" in value:
                function_detail = value["function"].split(":")
                if function_detail[0].startswith("numpy."):
                    function_path = function_detail[0]
                    function_name = None
                else:
                    # at this point function_path is only the file name.
                    function_path, function_name = function_detail
                    function_path = str(
                        self.citros.root_citros
                        / "parameter_setups"
                        / "functions"
                        / function_path
                    )

                function = load_function(function_path, function_name)
                # Check if function has a parameter named `context` and add it if so
                if value.get("args") is None:
                    value["args"] = []

                args = [evaluate_value(arg) for arg in value["args"]]

                try:
                    if (
                        function_name is not None
                        and "context" in inspect.signature(function).parameters
                    ):
                        args.append(context)
                except Exception as e:
                    self.log.exception(e)

                self.log.debug(f"function = {function_name}, args = {args}")
                result = function(*args)
                self.log.debug(
                    f"function = {function}, args = {args}, result = {result}"
                )

                # convert numpy scalars to native scalars, if needed.
                if isinstance(result, (numpy.generic)):
                    return result.item()
                return result

            # regular dictionary
            elif isinstance(value, dict):
                return {k: evaluate_value(v) for k, v in value.items()}

            # list
            elif isinstance(value, list):
                return [evaluate_value(v) for v in value]

            # the value is a string representing a reference to another key
            elif isinstance(value, str) and (
                value in all_keys or is_valid_reference_key(value)
            ):
                if value in visited_keys:
                    raise ValueError(f"Circular dependency detected with key: {value}")
                visited_keys.add(value)
                result = evaluate_value(recursive_key_search(self.data, value))
                visited_keys.remove(value)
                return result

            # any other type
            else:
                return value

        # the actual work - recursively evaluate the values in the given dictionary.
        return evaluate_value(self.data)

    def _save_rendered(self, config, destination):
        # callback running inside ROS workspace context.
        from ament_index_python.packages import get_package_share_directory

        for package_name, citros_config in config["packages"].items():
            self.log.debug(f"Saving config for [{package_name}]")

            path_to_package = None
            try:
                # get the path to the package install directory - the project must be sourced for it to work
                path_to_package = get_package_share_directory(package_name)
            except Exception as e:
                self.log.exception(e)
                continue

            if not path_to_package:
                continue

            path = Path(path_to_package, "config")

            # check if folder exists
            if not path.exists():
                self.log.debug(
                    f"No config directory {path} exits for pack:{package_name}. passing."
                )
                continue

            path = Path(path, "params.yaml")

            # check if file exists
            if not Path(path).exists():
                self.log.debug(
                    f"No config file {path} exits for package: {package_name}. passing."
                )
                continue

            with open(path, "r") as stream:
                try:
                    default_config = yaml.safe_load(stream)
                except yaml.YAMLError as exc:
                    self.log.exception(exc)

            # citros_config will overwrite default_config if the same key appears in both.
            merged_config = {**default_config, **citros_config}
            self.log.debug(json.dumps(merged_config, indent=4))

            # override default values
            with open(path, "w") as file:
                yaml.dump(merged_config, file)

            # sanity check
            if destination is None:
                raise ValueError(f"citros_params.save_config: {destination} is None.")

            # save for metadata
            Path(destination, "config").mkdir(exist_ok=True)
            with open(Path(destination, "config", f"{package_name}.yaml"), "w") as file:
                yaml.dump(merged_config, file)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from itertools import cycle
from sklearn.neural_network import MLPRegressor
from sklearn.mixture import GaussianMixture
from sklearn import preprocessing
from gmr import GMM
from numpy.typing import ArrayLike
from typing import Optional
import matplotlib.figure
from .citros_data import CitrosData
from citros.data.access._utils import _get_logger

class CitrosDataArray:
    """
    Store CitrosData objects in a "dbs" attribute for regression analysis.

    Parameters
    ----------
    dbs : list
        list of CitrosData objects
    log : log : logging.Logger, default None
        Logger to record log. If None, then the new logger is created.
    """

    _debug_flag = False

    def __init__(self, dbs = None, log = None):

        if log is None:
            self.log = _get_logger(__name__)
        if isinstance(dbs, list):
            self.dbs = dbs
        elif isinstance(dbs, CitrosData):
            self.dbs = [dbs]
        elif dbs is None:
            self.dbs = []
    
    def add_db(self, db: CitrosData):
        """
        Add one CitrosData object to CitrosDataArray.

        Parameters
        ----------
        db : analysis.citros_data.CitrosData
            CitrosData object to add to storage.
        """
        if isinstance(db, CitrosData):
            self.dbs.append(db)
        else:
            self.log.error('expected CitrosData object, but {} was given'.format(type(db)))
    
    def add_dbs(self, dbs: list):
        """
        Add list of CitrosData objects to CitrosDataArray.

        Parameters
        ----------
        dbs : list
            list of CitrosData objects to add to storage.
        """
        for db in dbs:
            self.add_db(db)

    def drop_db(self, value: int):
        """
        Remove CitrosData object from CitrosDataArray.

        If `value` is an int, then removes by index, 
        if `value` is a CitrosData object, then removes it if it exists in CitrosDataArray.

        Parameters
        ----------
        value : int or analysis.citros_data.CitrosData
            Object or index of object to remove.
        """
        if isinstance(value, int):
            try:
                self.dbs.pop(value)
            except IndexError:
                self.log.error('index is out of range')
        elif isinstance(value, CitrosData):
            try:
                self.dbs.remove(value)
            except ValueError:
                self.log.error('object is not in CitrosDataArray')

    def get_prediction(self, parameters: dict, method: str = 'poly', n_poly: int = 2, activation: str = 'relu', 
                       max_iter: int = 500, solver: str = 'lbfgs', hidden_layer_sizes: ArrayLike = (10,), 
                       alpha: float = 1e-16, fig: Optional[matplotlib.figure.Figure] = None, 
                       return_fig: bool = False, **kwargs):
        """
        Show the predictions based on the results of the regression solution, neural net or gaussian mixture model.

        Parameters
        ----------
        parameters : dict
            Names of the independent parameters and their values to calculate the prediction.
        method : str or list of str, default 'poly'
            - 'poly' - the polynomial regression.
            - 'neural_net' - the solution is finding based on [sklearn.neural_network.MLPRegressor](https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html).
            - 'gmm' - the gaussian mixture model is built and used for the prediction.
        n_poly : int, default 2
            Only used if `method` = 'poly'.
            The highest degree of the polynomial (1 for linear, 2 for quadratic, etc).
        activation : {'relu', 'identity', 'logistic' or 'tanh'}, default 'relu'
            Only used if `method` = 'neural_net'.
            Activation function for the hidden layer, see sklearn.neural_network.MLPRegressor
        max_iter : int, default 500
            Only used if `method` = 'neural_net'.
            Maximum number of iterations.
        solver : {'lbfgs', 'sgd', 'adam'}, default 'lbfgs'
            Only used if `method` = 'neural_net'.
            The solver for weight optimization.
        hidden_layer_sizes : array-like of shape(n_layers - 2,), default=(10,)
            Only used if `method` = 'neural_net'.
            The ith element represents the number of neurons in the ith hidden layer.
        alpha : float, default 1e-16
            Only used if `method` = 'gmm'.
            Value of the covariance element of parameters.
        fig : matplotlib.figure.Figure, optional
            figure to plot on. If None, then the new one is created.
        return_fig : bool, default False
            If True, the figure and ax (or list of ax) will be returned.

        Other Parameters
        ----------------
        kwargs : dict, optional
            Other keyword arguments for `method` = 'neural_net', see [sklearn.neural_network.MLPRegressor](https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html).

        Returns
        -------
        result : pandas.DataFrame
            Predicted table
        fig : matplotlib.figure.Figure
            if `return_fig` set to True
        ax : matplotlib.axes.Axes or list of matplotlib.axes.Axes
            if `return_fig` set to True

        Examples
        --------
        Create CitrosDataArray object:

        >>> from citros import CitrosDataArray, CitrosData
        >>> db_array = CitrosDataArray()

        Let's assume that for the topic 'A' from the batch 'star' of the simulation 'simulation_galaxy' we have simulations 
        for the four different values of the some parameter 't', that is written in json-data column 'data.t'. 
        To get list of the 'data.t' parameters get_unique_values() method may be used:

        >>> from citros import CitrosDB
        >>> citros = CitrosDB(simulation = 'simulation_galaxy', batch = 'star')
        >>> list_t = citros.topic('A').get_unique_values('data.t')
        >>> print(list_t)
        [-1.5, 0, 2.5, 4]

        Let's find prediction for the values of the 'data.x.x_1' json-column for the case when 'data.t' equals 1.
        Query data for each of these parameter values, set it as parameter, assign indexes over 'data.time' axis to set
        correspondence between different simulations and pass the result to CitrosDataArray that we created:

        >>> for t in list_t:
        ...     #query data
        ...     df = citros.topic('A')\\
        ...                .set_filter({'data.t': [t]})\\
        ...                .data(['data.x.x_1', 'data.time', 'data.t'])
        ...
        ...     #create CitrosData object and set 'data.t' as a parameter.
        ...     dataset = CitrosData(df,  
        ...                         data_label=['data.x.x_1'],
        ...                         units = 'm', 
        ...                         parameter_label = ['data.t'])
        ...
        ...     #scale over 'data.time'
        ...     db_sc = dataset.scale_data(n_points = 100, 
        ...                                param_label = 'data.time')
        ...
        ...     #store in CitrosDataArray by add_db() method
        ...     db_array.add_db(db_sc)

        Get the prediction with 'poly' method:

        >>> result = db_array.get_prediction(parameters = {'data.t': 1},
        ...                                  method = 'poly', 
        ...                                  n_poly = 2)
        >>> print(result)
            data.time	data.x.x_1
        0	0.000000	1.155301
        1	0.010101	1.145971
        2	0.020202	1.232255
        ...
        """
        slice_id_list = self.dbs[0].data.index.get_level_values(0).to_list()
        
        show_fig = True

        if not isinstance(method, list):
            method = [method]

        stat_df = pd.DataFrame()
        x_df = pd.DataFrame()
        cov_df = pd.DataFrame()
        for i in range(len(self.dbs)):
            stat = self.dbs[i].get_statistics(return_format = 'pandas')
            x_name = stat.columns[0]
            stat_df = pd.concat([stat_df, stat['mean'].rename('mean_'+str(i))], axis =1)
            cov_df = pd.concat([cov_df, stat['covar_matrix'].rename('cov_'+str(i))], axis =1)
            x_df = pd.concat([x_df, stat[self.dbs[i].x_label]], axis =1)
        stat_df.sort_index(inplace=True)
        x_df.sort_index(inplace=True)
        x_df = x_df.mean(axis = 1)
        N_ax = self.dbs[0].data_dim
        slice_id_list = stat_df.index.to_list()

        if show_fig or return_fig or fig is not None:
            if fig is None:
                fig, ax = plt.subplots(nrows = N_ax, ncols = 1,figsize=(6, 6))
                if not isinstance(ax, list):
                    ax = [ax]
            else:
                ax = []
                for i in range(1,N_ax+1):
                    ax.append(fig.add_subplot(N_ax, 1, i))
            for i in range(len(self.dbs)):
                filter_db = stat_df['mean_'+str(i)].notna()
                lab = ', '.join([p + ' = ' + str(self.dbs[i].parameters[p]) for p in parameters.keys()])
                for n in range(N_ax):
                    ax[n].plot(x_df[filter_db], stat_df['mean_'+str(i)][filter_db].apply(lambda x: x[n]), '-', linewidth = 0.9, label = lab)
                    ax[n].grid(True)

            lab_new = ', '.join([k + ' = ' + str(v) for k, v in parameters.items()])

            color_list = cycle(['k','darkred','rebeccapurple'])

        predicted_tables = []
        for method_step in method:
            mu_calculated = self._make_prediction(method_step, stat_df, cov_df, slice_id_list, n_poly, parameters,\
                                            N_ax, hidden_layer_sizes, activation, max_iter, solver, alpha, **kwargs)
            predicted_tables.append(pd.concat([x_df, pd.DataFrame(mu_calculated)], axis = 1))
            predicted_tables[-1].columns = [x_name] + list(self.dbs[0].data.columns)

            if show_fig or return_fig:
                color = next(color_list)
                for n in range(N_ax):
                    ax[n].plot(x_df, mu_calculated[:,n],'-', color = color, linewidth = 2, label = lab_new+',\n' + method_step)
                    ax[n].set_ylabel(self.dbs[0].data.columns[n]+', [' + self.dbs[0].units+']')

        if show_fig or return_fig or fig is not None:
            handles, labels = ax[-1].get_legend_handles_labels()
            fig.supxlabel(self.dbs[0].x_label)
            fig.suptitle('Prediction plot')
            fig.legend(handles, labels, bbox_to_anchor=(1.0, 0.94),loc ='upper left')
            fig.tight_layout()

        if len(predicted_tables) == 1:
            predicted_tables = predicted_tables[0]
        if return_fig:
            if len(ax) == 1:
                return predicted_tables, fig, ax[0]
            else:
                return predicted_tables, fig, ax
        else:
            return predicted_tables
    
    def _make_prediction(self, method, stat_df, cov_df, slice_id_list, n_poly, parameters, N_ax, hidden_layer_sizes,\
                        activation, max_iter, solver, alpha, **kwargs):
        """
        Inner function to calculate prediction for mean values by one of the methods.

        Parameters
        ----------
        method : str or list of str, default 'poly'
            - If the method is 'poly', the polynomial regression is solved.
            - If the method is 'neural_net', the solution is finding by sklearn.neural_network.MLPRegressor.
            - If the method is 'gmm', the gaussian mixture model is built and used for the prediction.
        stat_df : pandas.DataFrame
            DataFrame with prepared means values.
        cov_df : pandas.DataFrame
            DataFrame with prepared covariance matrixes.
        slice_id_list : list
            List of the indexes.
        n_poly : int, default 2
            Only used if method = 'poly'.
            The highest degree of the polynomial (1 for linear, 2 for quadratic, etc).
        parameters : dict
            Names of the independent parameters and their values to calculate the prediction.
        N_ax : int
            Number of the axis (equals dimension of the data).
        hidden_layer_sizes : array-like of shape(n_layers - 2,), default=(10,)
            Only used if method = 'neural_net'.
            The ith element represents the number of neurons in the ith hidden layer.
        activation : str, default 'relu'
            Only used if method = 'neural_net'.
            Activation function for the hidden layer, see sklearn.neural_network.MLPRegressor.
        max_iter : int, default 500
            Only used if method = 'neural_net'.
            Maximum number of iterations.
        solver : str, default 'lbfgs'
            Only used if method = 'neural_net'.
            The solver for weight optimization.
        alpha : float, default 1e-16
            Only used if method = 'gmm'.
            Value of the covariance element for parameters.
        **kwargs : dict, optional
            Other keyword arguments when method = 'neural_net', see [sklearn.neural_network.MLPRegressor](https://scikit-learn.org/stable/modules/generated/sklearn.neural_network.MLPRegressor.html).

        Returns
        -------
        out : numpy.ndarray
            Predicted mean values.
        """
        mu_calculated = []
        if method == 'poly':
            P_new = [1,] + [p**i for i in range(1, n_poly+1) for p in parameters.values()]
            P = np.array(list(map(lambda x: [x.parameters[p]**i for i in range(1, n_poly+1) for p in parameters.keys()], self.dbs)))
            P = np.append(np.array([[1]*len(self.dbs)]).T, P, axis = 1)
            for slice_id in slice_id_list:
                mu = np.array(stat_df.loc[slice_id].to_list())
                try:
                    mu_param = self._fit_regression(P, mu)
                    mu_calculated.append(self._calculate_regression(mu_param, P_new))
                except:
                    # mu_calculated.append([None]*N_ax if N_ax > 1 else None)
                    mu_calculated.append(np.array([np.nan]*N_ax))

        elif method == 'neural_net':
            P_new = np.array(list(parameters.values())).reshape(1,-1)
            P=np.array([[db.parameters[k] for k in parameters.keys()] for db in self.dbs])
            for slice_id in slice_id_list:
                mu = np.array(stat_df.loc[slice_id].to_list())
                pr = []
                try:
                    for j in range(mu.shape[1]):
                        nn = MLPRegressor(hidden_layer_sizes=hidden_layer_sizes, activation=activation, \
                            max_iter=max_iter, solver=solver, verbose=False, **kwargs)
                        mean = mu[:,j].mean()
                        std = mu[:,j].std()
                        try:
                            X = ((mu[:,j] - mean)/std)
                            n = nn.fit(P, X)
                            pr.append(nn.predict(P_new)[0]*std+mean)
                        except ZeroDivisionError:
                            n = nn.fit(P, mu[:,j])
                            pr.append(nn.predict(P_new)[0])
                    mu_calculated.append(np.array(pr))
                except Exception as e:
                    if CitrosDataArray._debug_flag:
                        raise e
                    mu_calculated.append(np.array([np.nan]*N_ax))

        elif method == 'gmm':
            P=np.array([[db.parameters[k] for k in parameters.keys()] for db in self.dbs])
            P_new = np.array(list(parameters.values())).reshape(-1,P.shape[1])
            for slice_id in slice_id_list:
                mu = np.array(stat_df.loc[slice_id].to_list())
                cov = np.array(cov_df.loc[slice_id].to_list())
                X = np.c_[mu, P]
                X_cov =[]
                for cov_element in cov:
                    X_cov.append(np.r_[np.c_[cov_element,np.ones((cov_element.shape[0],P.shape[1]))*alpha],np.ones((P.shape[1],cov_element.shape[1]+P.shape[1]))*alpha])
                X_cov = np.array(X_cov)
                try:
                    gmm_sklearn = GaussianMixture(n_components=len(self.dbs), covariance_type = 'full', verbose=False)
                    gmm_sklearn.fit(X)
                    gmm = GMM(
                        n_components=len(self.dbs), priors=gmm_sklearn.weights_, means=gmm_sklearn.means_,
                        covariances=X_cov)
                    pr_index = [i for i in range(N_ax,N_ax+P.shape[1])]
                    mu_calculated.append(gmm.predict(pr_index, P_new)[0])
                except Exception as e:
                    if CitrosDataArray._debug_flag:
                        raise e
                    # mu_calculated.append([None]*N_ax if N_ax > 1 else None)
                    mu_calculated.append(np.array([np.nan]*N_ax))
        else:
            self.log.error('there is no method called "{}". Try "poly", "neural_net" or "gmm".')
        return np.array(mu_calculated)
    
    def _data_for_regression(self, slice_id):
        """
        Prepare data for the regression fitting.

        Parameters
        ----------
        slice_id : int
            id at which the data is sliced.

        Returns
        -------
        mu : numpy.ndarray
            array of the mean values.
        sigma : numpy.ndarray
            array of the unique elements of the covariance matrix (diagonal and upper half).
        shape : tuple
            shape of the covariance matrix.
        """
        mu = []
        sigma = []
        for db in self.dbs:
            stat = db.get_statistics(return_format = 'pandas')
            mu.append(stat['mean'].loc[slice_id])
            covar = stat['covar_matrix'].loc[slice_id]
            s = []
            for i in range(len(covar)):
                for j in range(i, len(covar)):
                    s.append(stat['covar_matrix'].loc[slice_id][i,j])
            sigma.append(s)
        sigma = np.array(sigma)
        mu = np.array(mu)
        return mu, sigma, covar.shape

    def _fit_regression(self, P, y):
        """
        Fit the regression.

        Calculates the parameters x that solves y = Px.

        Parameters
        ----------
        P : numpy.ndarray
            Matrix of the coefficients.
        y : numpy.ndarray
            Vector of the dependent variables.

        Returns
        -------
        out : numpy.ndarray
            Vector of the regression parameters. 
        """
        x, res, rnk, s = np.linalg.lstsq(P, y, rcond = None) 
        return x

    def _get_regression_coef(self, slice_id, n_poly, parameters = None):
        """
        Calculate coefficients for the regression for mean values and for covariance matrix.

        Parameters
        ----------
        slice_id : int
            id at which the data is sliced.
        n_poly : int
            The highest degree of the polynomial (1 for linear, 2 for quadratic, etc)
        parameters : list
            Names of the independent parameters. If not specified, all parameters are used.

        Returns
        -------
        mu_param : numpy.ndarray
            Coefficients of the regression for mean value.
        sigma_param : numpy.ndarray
            Coefficients of the regression for covariance matrix.
            The shape is equals the shape of the covariance matrix.
        """
        if parameters is None:
            parameters = self.dbs[0].parameters.keys()
        mu, sigma, covar_shape = self._data_for_regression(self.dbs, slice_id)
        P = np.array(list(map(lambda x: [x.parameters[p]**i for i in range(1, n_poly+1) for p in parameters],self.dbs)))
        P = np.append(np.array([[1]*len(self.dbs)]).T, P, axis = 1)
        mu_param = self._fit_regression(P, mu) 
        res_sigma = self._fit_regression(P, sigma)
        
        sigma_param = []
        for res in res_sigma:
            k=0
            sigma_elem = np.empty(covar_shape)
            for i in range(covar_shape[0]):
                for j in range(i,covar_shape[0]):
                    sigma_elem[i,j] = res[k]
                    k+=1
                for j in range(i):
                    sigma_elem[i,j] = sigma_elem[j,i]
            sigma_param.append(sigma_elem)
            
        return mu_param, np.array(sigma_param)

    def _calculate_regression(self, x_param, P):
        """
        Calculate x_param * P.

        Function for checking the results of the regression algorithm.

        Parameters
        ----------
        x_param : array-like
            Parameters of the regression.
        P : array-like
            The vector of the coefficients.
        
        Returns
        -------
        out : float
            The result of the summing.
        """
        x_calc = 0
        for i in range(len(x_param)):
            x_calc = x_calc + P[i]*x_param[i]
        return x_calc


import json
import glob
import shutil
import importlib_resources

from os import linesep
from pathlib import Path

from citros.parsers import ParserRos2
from citros.utils import validate_dir, validate_file

# from .ros import Ros
from .settings import Settings
from .simulation import Simulation
from .parameter_setup import ParameterSetup
from .citros_obj import (
    CitrosObj,
    CitrosException,
    FileNotFoundException,
    CitrosNotFoundException,
    NoValidException,
)

from rich.traceback import install
from rich.logging import RichHandler
from rich import print, inspect, print_json
from rich.panel import Panel
from rich.padding import Padding
import shutil

install()

from .batch import Batch


class Citros(CitrosObj):
    """Object representing .citros/simulations/{name}.json file."""

    # def __enter__(self):
    #     """
    #     Returns the Citros instance. This allows the class to be used in a `with` statement.
    #     """
    #     return self

    # def __exit__(self, exc_type, exc_val, exc_tb):
    #     """
    #     Makes sure the stats collecting thread is stopped and handles exceptions.

    #     Args:
    #     exc_type: The type of exception.
    #     exc_val: The exception instance.
    #     exc_tb: A traceback object encapsulating the call stack at the point
    #             where the exception originally occurred.
    #     """
    #     self.events.on_shutdown()

    #     self.systemStatsRecorder.stop()

    #     if exc_type is not None:
    #         self._handle_exceptions(exc_val, exit=True)

    def __init__(
        self,
        name="project",
        root=None,
        new=False,
        log=None,
        citros=None,
        verbose=False,
        debug=False,
        level=0,
    ):
        ###################
        ##### .citros #####
        ###################
        # init settings
        self.settings = None

        # init parameter_setups
        self.parameter_setups = []

        # init simulations
        self.simulations = []

        #################
        ##### utils #####
        #################
        self._ros = None

        super().__init__(name, root, new, log, citros, verbose, debug, level)

    def __str__(self):
        # print_json(data=self.data)
        return json.dumps(self.data, indent=4)

    ###################
    ##### private #####
    ###################
    # overriding
    def _validate(self):
        """Validate the json file."""

        # TODO[critical]: check that the project.json file is valid

        return True

    # overriding
    def _load(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}.load()")

        # loads the main file (project.json)
        try:
            self.log.debug(f"loading .citros/project.json")
            super()._load()
        except FileNotFoundException as ex:
            self.log.error(f"simulation file {self.file} does not exist.")
            raise CitrosNotFoundException(
                f"simulation file {self.file} does not exist."
            )

        self.copy_default_citros_files()

        # loads the settings.json file
        self.log.debug(f"loading .citros/settings.json")
        self.settings = Settings(
            "settings",
            root=self.root,
            log=self.log,
            citros=self,
            new=self.new,
            debug=self.debug,
            verbose=self.verbose,
            level=self.level + 1,
        )

        # loads the parameter_setups
        for file in glob.glob(f"{self.root_citros}/parameter_setups/*.json"):
            file = file.split("/")[-1]
            self.log.debug(f"loading parameter_setup: {file}")
            self.parameter_setups.append(
                ParameterSetup(
                    file,
                    root=self.root,
                    new=self.new,
                    log=self.log,
                    citros=self,
                    debug=self.debug,
                    verbose=self.verbose,
                    level=self.level + 1,
                )
            )

        # loads the simulations
        for file in glob.glob(f"{self.root_citros}/simulations/*.json"):
            file = file.split("/")[-1]
            # self.simulations.append(Simulation(self.root, file, self.log, citros=self))
            self.log.debug(f"loading simulation: {file}")
            self.simulations.append(
                Simulation(
                    file,
                    root=self.root,
                    new=self.new,
                    log=self.log,
                    citros=self,
                    debug=self.debug,
                    verbose=self.verbose,
                    level=self.level + 1,
                )
            )

        # utils
        # self.ros = Ros(self.root, "ros.json", self.log)

    # overriding
    def _new(self):
        self.log.debug(f"{'   '*self.level}{self.__class__.__name__}._new()")

        # create the .citros folder
        Path(self.root_citros).mkdir(parents=True, exist_ok=True)
        # copy data files.
        self.copy_default_citros_files()

        # settings
        self.log.debug(f"creating .citros/settings.json")
        self.settings = Settings(
            "settings",
            root=self.root,
            log=self.log,
            citros=self,
            new=self.new,
            verbose=self.verbose,
            debug=self.debug,
            level=self.level + 1,
        )

        self._parser_ros2 = ParserRos2(self.log, self.get_citros_ignore_list())
        # get data from ros2
        project_data = self._parser_ros2.parse(str(self.root))
        with open(self.path(), "w") as file:
            json.dump(project_data, file, sort_keys=True, indent=4)

        self.data = project_data
        self._save()

        # parameter setups
        self.log.debug(f"creating .citros/parameter_setups/default_param_setup.json")
        self.parameter_setups = ParameterSetup(
            "default_param_setup",
            root=self.root,
            log=self.log,
            citros=self,
            new=self.new,
            verbose=self.verbose,
            debug=self.debug,
            level=self.level + 1,
        )

        # create simulation per launch file as default
        self.log.debug(f"creating .citros/simulations/*")
        self._create_simulations()

    #################
    ##### utils #####
    #################
    def _get_launches(self):
        """returns a list of launch objects

        Args:
            proj_json (Path): path to project.json file

        Returns:
            [{
                package: str,
                name: str
            }]: array of launch info
        """

        launch_info = []

        for package in self.citros.get("packages", []):
            for launch in package.get("launches", []):
                if "name" in launch:
                    launch_info.append(
                        {"package": package.get("name", ""), "name": launch["name"]}
                    )

        return launch_info

    def _create_simulations(self):
        launch_infos = self._get_launches()
        if not launch_infos:
            self.log.warning("No launch files found in user's project.")
            print(
                Panel.fit(
                    Padding(
                        "[yellow]No launch files found. [white]If you have launch files in your project, make sure they are of the form [green]*.launch.py ",
                        1,
                    ),
                    title="help",
                )
            )

            return

        # inspect(launch_infos)
        for launch in launch_infos:
            package_name = launch["package"]
            launch_file = launch["name"]

            # print("vova", self.root, f"simulation_{launch_file.split('.')[0]}.json")
            self.simulations.append(
                Simulation(
                    f"simulation_{launch_file.split('.')[0]}",
                    root=self.root,
                    new=self.new,
                    log=self.log,
                    citros=self,
                    package_name=package_name,
                    launch_file=launch_file,
                    verbose=self.verbose,
                    debug=self.debug,
                    level=self.level + 1,
                )
            )

    def get_citros_ignore_list(self):
        if Path(self.root_citros, ".citrosignore").exists():
            with open(Path(self.root_citros, ".citrosignore"), "r") as file:
                lines = [line.strip() for line in file if "#" not in line]
                self.log.debug(f".citrosignore contenrs: {lines}")
                return lines
        else:
            self.log.debug(f"Could not find .citrosignore in {self.root_citros}")
            return []

    def copy_default_citros_files(self):
        self.log.debug(
            f"{'   '*self.level}{self.__class__.__name__}.copy_default_citros_files()"
        )

        source_folder = importlib_resources.files("data.defaults.citros")

        self.log.debug(f"copying {source_folder} to {self.root_citros}")

        import os

        for root, dirs, files in os.walk(source_folder):
            relative_path = os.path.relpath(root, source_folder)
            destination_folder = os.path.join(self.root_citros, relative_path)
            os.makedirs(destination_folder, exist_ok=True)
            for file in files:
                source_file = os.path.join(root, file)
                destination_file = os.path.join(destination_folder, file)
                if not os.path.exists(destination_file):
                    shutil.copy2(source_file, destination_file)

        self.log.debug(f"Done copying default Citros files.")

    ###################
    ##### public ######
    ###################
    def delete_batch(self, simulation: str, name: str, version: str):
        import shutil

        shutil.rmtree(self.root_citros / "data" / simulation / name / version)

    def get_batch(self, simulation: str, name: str, version=-1):
        batch = Batch(
            self.root_citros / "data",
            simulation,
            name=name,
            version=version if type(version) is str else None,
            index=version if type(version) is int else None,
            debug=self.debug,
            verbose=self.verbose,
        )

        return batch

    def unload_batch(self, simulation, batch):
        versions = glob.glob(f"{simulation}/{batch}/*")
        for ver in versions:
            batch = self.get_batch(simulation, batch, ver)
            batch.unload()

    def get_batches(self, simulation=None, batch=None, filter: str = None):
        batches = []
        if simulation:
            simulations = sorted(
                glob.glob(f"{str(self.root_citros / 'data' / simulation)}")
            )
        else:
            simulations = sorted(glob.glob(f"{str(self.root_citros / 'data')}/*/"))

        for sim in simulations:
            if batch:
                batch_names = sorted(glob.glob(f"{sim}/{batch}/"))
            else:
                batch_names = sorted(glob.glob(f"{sim}/*/"))

            for batch_name in batch_names:
                versions = sorted(glob.glob(f"{batch_name}/*/"), reverse=True)

                batches.append(
                    {
                        "simulation": (sim if sim[-1] != "/" else sim[:-1]).split("/")[
                            -1
                        ],
                        "name": (
                            batch_name if batch_name[-1] != "/" else batch_name[:-1]
                        ).split("/")[-1],
                        "versions": [
                            (v if v[-1] != "/" else v[:-1]).split("/")[-1]
                            for v in versions
                        ],
                    }
                )

        return batches

    def get_batches_flat(self):
        batches = []
        simulations = sorted(glob.glob(f"{str(self.root_citros / 'data')}/*/"))
        for sim in simulations:
            if not Path(sim).is_dir():
                continue
            names = sorted(glob.glob(f"{sim}/*/"))
            _simulation = (sim if sim[-1] != "/" else sim[:-1]).split("/")[-1]
            for name in names:
                if not Path(name).is_dir():
                    continue
                versions = sorted(glob.glob(f"{name}/*/"), reverse=True)
                # print(versions)
                _name = (name if name[-1] != "/" else name).split("/")[-1]
                try:
                    batch_hr_statu = json.loads((Path(name) / "hr.json").read_text())
                except:
                    batch_hr_statu = {"status": "UNLOADED"}

                for version in versions:
                    batch = json.loads((Path(version) / "info.json").read_text())
                    status = "UNLOADED"
                    if (
                        batch_hr_statu.get("version", None)
                        == (version if version[-1] != "/" else version).split("/")[-1]
                    ):
                        status = batch_hr_statu.get("status")

                    batches.append(
                        {
                            "created_at": batch["created_at"],
                            "simulation": _simulation,
                            "name": _name,
                            "version": (
                                version if version[-1] != "/" else version
                            ).split("/")[-1],
                            "message": batch["message"],
                            f"status": status,
                            "completions": str(batch["completions"]),
                            "path": version,
                        }
                    )

                    # for printing.
                    _simulation = None
                    _name = None

        return batches

    def get_reports_flat(self):
        ret = []
        reports = sorted(glob.glob(f"{str(self.root_citros / 'reports')}/*/"))
        for report in reports:
            if not Path(report).is_dir():
                continue
            versions = sorted(glob.glob(f"{report}/*/"), reverse=True)

            for version in versions:
                report_version = json.loads((Path(version) / "info.json").read_text())

                notebooks = ""
                if (
                    report_version.get("notebooks") == None
                    or report_version["notebooks"] == None
                    or len(report_version["notebooks"]) == 0
                ):
                    continue
                else:
                    report_version_notebook = (
                        report_version["notebooks"][0]
                        if report_version["notebooks"][0][-1] != "/"
                        else report_version["notebooks"][0][:-1]
                    )
                    notebooks = str(
                        Path(version) / report_version_notebook.split("/")[-1]
                    )
                    if notebooks.endswith(".ipynb"):
                        notebooks = notebooks[: -len(".ipynb")]

                ret.append(
                    {
                        "started_at": report_version["started_at"],
                        "finished_at": report_version["finished_at"],
                        "name": report_version["name"],
                        "version": (version if version[-1] != "/" else version).split(
                            "/"
                        )[-1],
                        "message": report_version["message"],
                        "path": f'{str(self.root_citros / "reports" / report_version["name"]/ (version if version[-1] != "/" else version).split("/")[-1])}',
                        "progress": report_version["progress"],
                        "status": report_version["status"],
                    }
                )

        return ret
